from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class InvalidContainerRegistryProject(BackendAIError, web.HTTPBadRequest):
    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.CONTAINER_REGISTRY,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class InvalidContainerRegistryURL(BackendAIError, web.HTTPBadRequest):
    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.CONTAINER_REGISTRY,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.BAD_REQUEST,
        )



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.resource_preset.actions.check_presets import (
    CheckResourcePresetsAction,
    CheckResourcePresetsActionResult,
)
from ai.backend.manager.services.resource_preset.actions.create_preset import (
    CreateResourcePresetAction,
    CreateResourcePresetActionResult,
)
from ai.backend.manager.services.resource_preset.actions.delete_preset import (
    DeleteResourcePresetAction,
    DeleteResourcePresetActionResult,
)
from ai.backend.manager.services.resource_preset.actions.list_presets import (
    ListResourcePresetsAction,
    ListResourcePresetsResult,
)
from ai.backend.manager.services.resource_preset.actions.modify_preset import (
    ModifyResourcePresetAction,
    ModifyResourcePresetActionResult,
)
from ai.backend.manager.services.resource_preset.service import ResourcePresetService


class ResourcePresetProcessors(AbstractProcessorPackage):
    create_preset: ActionProcessor[CreateResourcePresetAction, CreateResourcePresetActionResult]
    modify_preset: ActionProcessor[ModifyResourcePresetAction, ModifyResourcePresetActionResult]
    delete_preset: ActionProcessor[DeleteResourcePresetAction, DeleteResourcePresetActionResult]
    list_presets: ActionProcessor[ListResourcePresetsAction, ListResourcePresetsResult]
    check_presets: ActionProcessor[CheckResourcePresetsAction, CheckResourcePresetsActionResult]

    def __init__(
        self, service: ResourcePresetService, action_monitors: list[ActionMonitor]
    ) -> None:
        self.create_preset = ActionProcessor(service.create_preset, action_monitors)
        self.modify_preset = ActionProcessor(service.modify_preset, action_monitors)
        self.delete_preset = ActionProcessor(service.delete_preset, action_monitors)
        self.list_presets = ActionProcessor(service.list_presets, action_monitors)
        self.check_presets = ActionProcessor(service.check_presets, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            CreateResourcePresetAction.spec(),
            ModifyResourcePresetAction.spec(),
            DeleteResourcePresetAction.spec(),
            ListResourcePresetsAction.spec(),
            CheckResourcePresetsAction.spec(),
        ]



import logging

from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.common.events.event_types.artifact.anycast import (
    ModelImportDoneEvent,
    ModelMetadataFetchDoneEvent,
)
from ai.backend.common.types import (
    AgentId,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.artifact.types import ArtifactStatus
from ai.backend.manager.errors.artifact_registry import InvalidArtifactRegistryTypeError
from ai.backend.manager.repositories.artifact.repository import ArtifactRepository
from ai.backend.manager.repositories.huggingface_registry.repository import HuggingFaceRepository
from ai.backend.manager.repositories.reservoir_registry.repository import (
    ReservoirRegistryRepository,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ArtifactEventHandler:
    _artifact_repository: ArtifactRepository
    _huggingface_repository: HuggingFaceRepository
    _reservoir_repository: ReservoirRegistryRepository
    _config_provider: ManagerConfigProvider

    def __init__(
        self,
        artifact_repository: ArtifactRepository,
        huggingface_repository: HuggingFaceRepository,
        reservoir_repository: ReservoirRegistryRepository,
        config_provider: ManagerConfigProvider,
    ) -> None:
        self._artifact_repository = artifact_repository
        self._huggingface_repository = huggingface_repository
        self._reservoir_repository = reservoir_repository
        self._config_provider = config_provider

    async def handle_model_import_done(
        self,
        context: None,
        source: AgentId,
        event: ModelImportDoneEvent,
    ) -> None:
        try:
            registry_type = ArtifactRegistryType(event.registry_type)
        except Exception:
            raise InvalidArtifactRegistryTypeError(
                f"Unsupported artifact registry type: {event.registry_type}"
            )

        match registry_type:
            case ArtifactRegistryType.HUGGINGFACE:
                registry_data = await self._huggingface_repository.get_registry_data_by_name(
                    event.registry_name
                )
            case ArtifactRegistryType.RESERVOIR:
                registry_data = await self._reservoir_repository.get_registry_data_by_name(
                    event.registry_name
                )

        artifact = await self._artifact_repository.get_model_artifact(
            event.model_id, registry_id=registry_data.id
        )

        # Get the specific revision
        revision = await self._artifact_repository.get_artifact_revision(
            artifact.id, revision=event.revision
        )

        try:
            if self._config_provider.config.reservoir.enable_approve_process:
                await self._artifact_repository.update_artifact_revision_status(
                    revision.id, ArtifactStatus.NEEDS_APPROVAL
                )
            else:
                await self._artifact_repository.update_artifact_revision_status(
                    revision.id, ArtifactStatus.AVAILABLE
                )
        except Exception as model_error:
            log.error(
                "Failed to process imported model update: {} - {}",
                event.model_id,
                model_error,
            )

    async def handle_model_metadata_fetch_done(
        self,
        context: None,
        source: AgentId,
        event: ModelMetadataFetchDoneEvent,
    ) -> None:
        model_info = event.model
        try:
            registry_type = ArtifactRegistryType(model_info.registry_type)
        except Exception:
            raise InvalidArtifactRegistryTypeError(
                f"Unsupported artifact registry type: {model_info.registry_type}"
            )

        match registry_type:
            case ArtifactRegistryType.HUGGINGFACE:
                registry_data = await self._huggingface_repository.get_registry_data_by_name(
                    model_info.registry_name
                )
            case ArtifactRegistryType.RESERVOIR:
                registry_data = await self._reservoir_repository.get_registry_data_by_name(
                    model_info.registry_name
                )

        artifact = await self._artifact_repository.get_model_artifact(
            model_info.model_id, registry_id=registry_data.id
        )

        # Get the specific revision
        revision = await self._artifact_repository.get_artifact_revision(
            artifact.id, revision=model_info.revision
        )

        try:
            # Update the README content
            await self._artifact_repository.update_artifact_revision_readme(
                revision.id, model_info.readme_content
            )

            # Update the file size
            await self._artifact_repository.update_artifact_revision_bytesize(
                revision.id, model_info.size
            )

            log.trace(
                "Updated metadata for model: {} revision: {} in artifact: {} (size: {} bytes)",
                model_info.model_id,
                model_info.revision,
                artifact.id,
                model_info.size,
            )
        except Exception as model_error:
            log.error(
                "Failed to process metadata update for model: {} - {}",
                model_info.model_id,
                model_error,
            )



"""Endpoint and route data types for deployment repository."""

import uuid
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Optional

from ai.backend.common.types import SessionId
from ai.backend.manager.data.model_serving.types import EndpointLifecycle, RouteStatus


@dataclass
class EndpointCreationArgs:
    """Arguments for creating an endpoint."""

    name: str
    model_id: uuid.UUID
    owner_id: uuid.UUID
    group_id: uuid.UUID
    domain_name: str
    is_public: bool
    runtime_variant: str
    desired_session_count: int
    resource_opts: Optional[dict[str, Any]] = None
    scaling_group: Optional[str] = None


@dataclass
class EndpointData:
    """Data structure for model service endpoint."""

    endpoint_id: uuid.UUID
    name: str
    model_id: uuid.UUID
    owner_id: uuid.UUID
    group_id: uuid.UUID
    domain_name: str
    lifecycle: EndpointLifecycle
    is_public: bool
    runtime_variant: str
    desired_session_count: int
    created_at: datetime
    updated_at: Optional[datetime] = None
    service_endpoint: Optional[str] = None
    resource_opts: dict[str, Any] = field(default_factory=dict)


@dataclass
class RouteData:
    """Data structure for model service route."""

    route_id: uuid.UUID
    endpoint_id: uuid.UUID
    session_id: Optional[SessionId]
    status: RouteStatus
    traffic_ratio: float
    created_at: datetime
    updated_at: Optional[datetime] = None
    error_data: dict[str, Any] = field(default_factory=dict)



"""add scaling_group_name column to resource_presets table

Revision ID: d537c050dd8d
Revises: c002140f14d3
Create Date: 2025-02-16 18:20:18.694834

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "d537c050dd8d"
down_revision = "c002140f14d3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Add scaling_group_name column
    op.add_column(
        "resource_presets",
        sa.Column(
            "scaling_group_name",
            sa.String(length=64),
            server_default=sa.text("NULL"),
            nullable=True,
        ),
    )

    # Migrate primary key from `name` to `id`
    op.drop_constraint(op.f("pk_resource_presets"), "resource_presets", type_="primary")
    op.add_column(
        "resource_presets",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
    )
    op.create_primary_key("pk_resource_presets", "resource_presets", ["id"])


def downgrade() -> None:
    op.drop_column("resource_presets", "scaling_group_name")

    op.drop_column("resource_presets", "id")
    # The resource_presets.name column lacks a unique/primary key constraint. Creation
    # of a primary key may fail if duplicate names exist. Manually resolve any duplicate
    # resource_preset names before retrying.
    op.create_primary_key("pk_resource_presets", "resource_presets", ["name"])



import uuid
from collections.abc import Mapping
from dataclasses import dataclass
from typing import (
    Any,
    Optional,
    override,
)

from ai.backend.manager.actions.action import BaseAction, BaseActionResult
from ai.backend.manager.models.vfolder import VFolderPermission as VFolderMountPermission

from ..types import VFolderInvitationInfo


@dataclass
class VFolderInvitationAction(BaseAction):
    @override
    @classmethod
    def entity_type(cls) -> str:
        return "vfolder_invitation"


@dataclass
class InviteVFolderAction(VFolderInvitationAction):
    keypair_resource_policy: Mapping[str, Any]
    user_uuid: uuid.UUID

    vfolder_uuid: uuid.UUID
    mount_permission: VFolderMountPermission
    invitee_user_uuids: list[uuid.UUID]

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "invite"


@dataclass
class InviteVFolderActionResult(BaseActionResult):
    vfolder_uuid: uuid.UUID
    invitation_ids: list[uuid.UUID]

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)


@dataclass
class AcceptInvitationAction(VFolderInvitationAction):
    invitation_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.invitation_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "accept"


@dataclass
class AcceptInvitationActionResult(BaseActionResult):
    invitation_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.invitation_id)


@dataclass
class RejectInvitationAction(VFolderInvitationAction):
    invitation_id: uuid.UUID
    requester_user_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.invitation_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "reject"


@dataclass
class RejectInvitationActionResult(BaseActionResult):
    invitation_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.invitation_id)


@dataclass
class UpdateInvitationAction(VFolderInvitationAction):
    invitation_id: uuid.UUID

    requester_user_uuid: uuid.UUID
    mount_permission: VFolderMountPermission

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.invitation_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "update"


@dataclass
class UpdateInvitationActionResult(BaseActionResult):
    invitation_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.invitation_id)


@dataclass
class ListInvitationAction(VFolderInvitationAction):
    requester_user_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.requester_user_uuid)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "list"


@dataclass
class ListInvitationActionResult(BaseActionResult):
    requester_user_uuid: uuid.UUID
    info: list[VFolderInvitationInfo]

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.requester_user_uuid)


@dataclass
class LeaveInvitedVFolderAction(VFolderInvitationAction):
    vfolder_uuid: uuid.UUID
    requester_user_uuid: uuid.UUID
    shared_user_uuid: Optional[uuid.UUID] = None

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "leave"


@dataclass
class LeaveInvitedVFolderActionResult(BaseActionResult):
    vfolder_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)


@dataclass
class RevokeInvitedVFolderAction(VFolderInvitationAction):
    vfolder_id: uuid.UUID
    shared_user_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "revoke"


@dataclass
class RevokeInvitedVFolderActionResult(BaseActionResult):
    vfolder_id: uuid.UUID
    shared_user_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_id)


@dataclass
class UpdateInvitedVFolderMountPermissionAction(VFolderInvitationAction):
    vfolder_id: uuid.UUID
    user_id: uuid.UUID
    permission: VFolderMountPermission

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "update_permission"


@dataclass
class UpdateInvitedVFolderMountPermissionActionResult(BaseActionResult):
    vfolder_id: uuid.UUID
    user_id: uuid.UUID
    permission: VFolderMountPermission

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_id)



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Optional, Self

import sqlalchemy as sa
from sqlalchemy.orm import relationship

from ai.backend.manager.data.permission.id import ScopeId
from ai.backend.manager.data.permission.permission_group import (
    PermissionGroupCreator,
    PermissionGroupData,
)
from ai.backend.manager.data.permission.types import (
    ScopeType,
)

from ...base import (
    GUID,
    Base,
    IDColumn,
    StrEnumType,
)

if TYPE_CHECKING:
    from ..role import RoleRow
    from .permission import PermissionRow


class PermissionGroupRow(Base):
    __tablename__ = "permission_groups"
    __table_args__ = (sa.Index("ix_id_role_id_scope_id", "id", "role_id", "scope_id"),)

    id: uuid.UUID = IDColumn()
    role_id: uuid.UUID = sa.Column("role_id", GUID, nullable=False)
    scope_type: ScopeType = sa.Column(
        "scope_type", StrEnumType(ScopeType, length=32), nullable=False
    )
    scope_id: str = sa.Column(
        "scope_id", sa.String(64), nullable=False
    )  # e.g., "project_id", "user_id" etc.

    role_row: Optional[RoleRow] = relationship(
        "RoleRow",
        back_populates="permission_group_rows",
        primaryjoin="RoleRow.id == foreign(PermissionGroupRow.role_id)",
    )
    permission_rows: list[PermissionRow] = relationship(
        "PermissionRow",
        back_populates="permission_group_row",
        primaryjoin="PermissionGroupRow.id == foreign(PermissionRow.permission_group_id)",
    )

    def parsed_scope_id(self) -> ScopeId:
        return ScopeId(scope_type=self.scope_type, scope_id=self.scope_id)

    @classmethod
    def from_input(cls, input: PermissionGroupCreator) -> Self:
        return cls(
            role_id=input.role_id,
            scope_type=input.scope_id.scope_type,
            scope_id=input.scope_id.scope_id,
        )

    def to_data(self) -> PermissionGroupData:
        return PermissionGroupData(
            id=self.id,
            role_id=self.role_id,
            scope_id=ScopeId(scope_type=self.scope_type, scope_id=self.scope_id),
        )



from dataclasses import dataclass, field
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.user.types import UserInfoContext
from ai.backend.manager.services.user.actions.base import UserAction
from ai.backend.manager.types import OptionalState


@dataclass
class PurgeUserAction(UserAction):
    user_info_ctx: UserInfoContext
    email: str
    purge_shared_vfolders: OptionalState[bool] = field(default_factory=OptionalState.nop)
    delegate_endpoint_ownership: OptionalState[bool] = field(default_factory=OptionalState.nop)

    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "purge"


@dataclass
class PurgeUserActionResult(BaseActionResult):
    def entity_id(self) -> Optional[str]:
        return None



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.domain.types import (
    DomainCreator,
    DomainData,
    UserInfo,
)
from ai.backend.manager.services.domain.actions.base import DomainAction


@dataclass
class CreateDomainNodeAction(DomainAction):
    user_info: UserInfo
    creator: DomainCreator
    scaling_groups: Optional[list[str]] = None

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "create"


@dataclass
class CreateDomainNodeActionResult(BaseActionResult):
    domain_data: DomainData

    @override
    def entity_id(self) -> Optional[str]:
        return self.domain_data.name



"""add model_definition_path column

Revision ID: f56a82d0ac9f
Revises: c8714799ca0e
Create Date: 2024-04-24 23:17:02.139270

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f56a82d0ac9f"
down_revision = "c8714799ca0e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "endpoints", sa.Column("model_definition_path", sa.String(length=512), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("endpoints", "model_definition_path")
    # ### end Alembic commands ###



from dataclasses import dataclass
from typing import Optional, Self, override

from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import EventDispatcher, EventProducer
from ai.backend.common.events.fetcher import EventFetcher
from ai.backend.common.events.hub.hub import EventHub
from ai.backend.common.plugin.hook import HookPluginContext
from ai.backend.common.plugin.monitor import ErrorPluginContext
from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.agent_cache import AgentRPCCache
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.idle import IdleCheckerHost
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.repositories.repositories import Repositories
from ai.backend.manager.services.agent.processors import AgentProcessors
from ai.backend.manager.services.agent.service import AgentService
from ai.backend.manager.services.artifact.processors import ArtifactProcessors
from ai.backend.manager.services.artifact.service import ArtifactService
from ai.backend.manager.services.artifact_registry.processors import ArtifactRegistryProcessors
from ai.backend.manager.services.artifact_registry.service import ArtifactRegistryService
from ai.backend.manager.services.artifact_revision.processors import ArtifactRevisionProcessors
from ai.backend.manager.services.artifact_revision.service import ArtifactRevisionService
from ai.backend.manager.services.auth.processors import AuthProcessors
from ai.backend.manager.services.auth.service import AuthService
from ai.backend.manager.services.container_registry.processors import ContainerRegistryProcessors
from ai.backend.manager.services.container_registry.service import ContainerRegistryService
from ai.backend.manager.services.deployment.processors import DeploymentProcessors
from ai.backend.manager.services.deployment.service import DeploymentService
from ai.backend.manager.services.domain.processors import DomainProcessors
from ai.backend.manager.services.domain.service import DomainService
from ai.backend.manager.services.group.processors import GroupProcessors
from ai.backend.manager.services.group.service import GroupService
from ai.backend.manager.services.image.processors import ImageProcessors
from ai.backend.manager.services.image.service import ImageService
from ai.backend.manager.services.keypair_resource_policy.processors import (
    KeypairResourcePolicyProcessors,
)
from ai.backend.manager.services.keypair_resource_policy.service import KeypairResourcePolicyService
from ai.backend.manager.services.metric.processors.utilization_metric import (
    UtilizationMetricProcessors,
)
from ai.backend.manager.services.metric.root_service import UtilizationMetricService
from ai.backend.manager.services.model_serving.processors.auto_scaling import (
    ModelServingAutoScalingProcessors,
)
from ai.backend.manager.services.model_serving.processors.model_serving import (
    ModelServingProcessors,
    ModelServingServiceProtocol,
)
from ai.backend.manager.services.model_serving.services.auto_scaling import AutoScalingService
from ai.backend.manager.services.model_serving.services.model_serving import (
    ModelServingService,
)
from ai.backend.manager.services.object_storage.processors import ObjectStorageProcessors
from ai.backend.manager.services.object_storage.service import ObjectStorageService
from ai.backend.manager.services.project_resource_policy.processors import (
    ProjectResourcePolicyProcessors,
)
from ai.backend.manager.services.project_resource_policy.service import ProjectResourcePolicyService
from ai.backend.manager.services.resource_preset.processors import ResourcePresetProcessors
from ai.backend.manager.services.resource_preset.service import ResourcePresetService
from ai.backend.manager.services.session.processors import SessionProcessors
from ai.backend.manager.services.session.service import SessionService, SessionServiceArgs
from ai.backend.manager.services.storage_namespace.processors import StorageNamespaceProcessors
from ai.backend.manager.services.storage_namespace.service import StorageNamespaceService
from ai.backend.manager.services.user.processors import UserProcessors
from ai.backend.manager.services.user.service import UserService
from ai.backend.manager.services.user_resource_policy.processors import UserResourcePolicyProcessors
from ai.backend.manager.services.user_resource_policy.service import UserResourcePolicyService
from ai.backend.manager.services.vfolder.processors import (
    VFolderFileProcessors,
    VFolderInviteProcessors,
    VFolderProcessors,
)
from ai.backend.manager.services.vfolder.services.file import VFolderFileService
from ai.backend.manager.services.vfolder.services.invite import VFolderInviteService
from ai.backend.manager.services.vfolder.services.vfolder import VFolderService
from ai.backend.manager.sokovan.deployment import DeploymentController
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController


@dataclass
class ServiceArgs:
    db: ExtendedAsyncSAEngine
    repositories: Repositories
    etcd: AsyncEtcd
    config_provider: ManagerConfigProvider
    storage_manager: StorageSessionManager
    valkey_stat_client: ValkeyStatClient
    valkey_live: ValkeyLiveClient
    event_fetcher: EventFetcher
    background_task_manager: BackgroundTaskManager
    event_hub: EventHub
    agent_registry: AgentRegistry
    error_monitor: ErrorPluginContext
    idle_checker_host: IdleCheckerHost
    event_dispatcher: EventDispatcher
    hook_plugin_ctx: HookPluginContext
    scheduling_controller: SchedulingController
    deployment_controller: DeploymentController
    event_producer: EventProducer
    agent_cache: AgentRPCCache


@dataclass
class Services:
    agent: AgentService
    domain: DomainService
    group: GroupService
    user: UserService
    image: ImageService
    container_registry: ContainerRegistryService
    vfolder: VFolderService
    vfolder_file: VFolderFileService
    vfolder_invite: VFolderInviteService
    session: SessionService
    keypair_resource_policy: KeypairResourcePolicyService
    user_resource_policy: UserResourcePolicyService
    project_resource_policy: ProjectResourcePolicyService
    resource_preset: ResourcePresetService
    utilization_metric: UtilizationMetricService
    model_serving: ModelServingServiceProtocol
    model_serving_auto_scaling: AutoScalingService
    auth: AuthService
    object_storage: ObjectStorageService
    artifact: ArtifactService
    artifact_revision: ArtifactRevisionService
    artifact_registry: ArtifactRegistryService
    deployment: DeploymentService
    storage_namespace: StorageNamespaceService

    @classmethod
    def create(cls, args: ServiceArgs) -> Self:
        repositories = args.repositories
        agent_service = AgentService(
            args.etcd,
            args.agent_registry,
            args.config_provider,
            repositories.agent.repository,
            repositories.scheduler.repository,
            args.hook_plugin_ctx,
            args.event_producer,
            args.agent_cache,
        )
        domain_service = DomainService(
            repositories.domain.repository, repositories.domain.admin_repository
        )
        group_service = GroupService(
            args.storage_manager,
            args.config_provider,
            args.valkey_stat_client,
            repositories.group,
        )
        user_service = UserService(
            args.storage_manager,
            args.valkey_stat_client,
            args.agent_registry,
            repositories.user.repository,
            repositories.user.admin_repository,
        )
        image_service = ImageService(
            args.agent_registry, repositories.image.repository, repositories.image.admin_repository
        )
        container_registry_service = ContainerRegistryService(
            args.db,
            repositories.container_registry.repository,
            repositories.container_registry.admin_repository,
        )
        vfolder_service = VFolderService(
            args.config_provider,
            args.storage_manager,
            args.background_task_manager,
            repositories.vfolder.repository,
            repositories.user.repository,
        )
        vfolder_file_service = VFolderFileService(
            args.config_provider,
            args.storage_manager,
            repositories.vfolder.repository,
            repositories.user.repository,
        )
        vfolder_invite_service = VFolderInviteService(
            args.config_provider,
            repositories.vfolder.repository,
            repositories.user.repository,
        )
        session_service = SessionService(
            SessionServiceArgs(
                agent_registry=args.agent_registry,
                event_fetcher=args.event_fetcher,
                background_task_manager=args.background_task_manager,
                event_hub=args.event_hub,
                error_monitor=args.error_monitor,
                idle_checker_host=args.idle_checker_host,
                session_repository=repositories.session.repository,
                admin_session_repository=repositories.session.admin_repository,
                scheduling_controller=args.scheduling_controller,
            )
        )
        keypair_resource_policy_service = KeypairResourcePolicyService(
            repositories.keypair_resource_policy.repository
        )
        user_resource_policy_service = UserResourcePolicyService(
            repositories.user_resource_policy.repository
        )
        project_resource_policy_service = ProjectResourcePolicyService(
            repositories.project_resource_policy.repository
        )
        resource_preset_service = ResourcePresetService(
            repositories.resource_preset.repository,
        )
        utilization_metric_service = UtilizationMetricService(
            args.config_provider, repositories.metric.repository
        )

        # Use deployment-based model serving if deployment_controller is available
        model_serving_service = ModelServingService(
            agent_registry=args.agent_registry,
            background_task_manager=args.background_task_manager,
            event_dispatcher=args.event_dispatcher,
            storage_manager=args.storage_manager,
            config_provider=args.config_provider,
            valkey_live=args.valkey_live,
            repository=repositories.model_serving.repository,
            admin_repository=repositories.model_serving.admin_repository,
            deployment_controller=args.deployment_controller,
        )

        model_serving_auto_scaling = AutoScalingService(
            repository=repositories.model_serving.repository,
            admin_repository=repositories.model_serving.admin_repository,
        )
        auth = AuthService(
            hook_plugin_ctx=args.hook_plugin_ctx,
            auth_repository=repositories.auth.repository,
            config_provider=args.config_provider,
        )
        object_storage_service = ObjectStorageService(
            artifact_repository=repositories.artifact.repository,
            object_storage_repository=repositories.object_storage.repository,
            storage_manager=args.storage_manager,
            config_provider=args.config_provider,
        )
        artifact_service = ArtifactService(
            artifact_repository=repositories.artifact.repository,
            artifact_registry_repository=repositories.artifact_registry.repository,
            storage_manager=args.storage_manager,
            object_storage_repository=repositories.object_storage.repository,
            huggingface_registry_repository=repositories.huggingface_registry.repository,
            config_provider=args.config_provider,
            reservoir_registry_repository=repositories.reservoir_registry.repository,
        )
        artifact_revision_service = ArtifactRevisionService(
            artifact_repository=repositories.artifact.repository,
            storage_manager=args.storage_manager,
            object_storage_repository=repositories.object_storage.repository,
            huggingface_registry_repository=repositories.huggingface_registry.repository,
            reservoir_registry_repository=repositories.reservoir_registry.repository,
            config_provider=args.config_provider,
        )
        artifact_registry_service = ArtifactRegistryService(
            repositories.huggingface_registry.repository,
            repositories.reservoir_registry.repository,
            repositories.artifact_registry.repository,
        )
        deployment_service = DeploymentService(args.deployment_controller)
        storage_namespace_service = StorageNamespaceService(
            repositories.storage_namespace.repository
        )

        return cls(
            agent=agent_service,
            domain=domain_service,
            group=group_service,
            user=user_service,
            image=image_service,
            container_registry=container_registry_service,
            vfolder=vfolder_service,
            vfolder_file=vfolder_file_service,
            vfolder_invite=vfolder_invite_service,
            session=session_service,
            keypair_resource_policy=keypair_resource_policy_service,
            user_resource_policy=user_resource_policy_service,
            project_resource_policy=project_resource_policy_service,
            resource_preset=resource_preset_service,
            utilization_metric=utilization_metric_service,
            model_serving=model_serving_service,
            model_serving_auto_scaling=model_serving_auto_scaling,
            auth=auth,
            object_storage=object_storage_service,
            artifact=artifact_service,
            artifact_revision=artifact_revision_service,
            artifact_registry=artifact_registry_service,
            deployment=deployment_service,
            storage_namespace=storage_namespace_service,
        )


@dataclass
class ProcessorArgs:
    service_args: ServiceArgs


@dataclass
class Processors(AbstractProcessorPackage):
    agent: AgentProcessors
    domain: DomainProcessors
    group: GroupProcessors
    user: UserProcessors
    image: ImageProcessors
    vfolder: VFolderProcessors
    vfolder_invite: VFolderInviteProcessors
    vfolder_file: VFolderFileProcessors
    session: SessionProcessors
    container_registry: ContainerRegistryProcessors
    keypair_resource_policy: KeypairResourcePolicyProcessors
    user_resource_policy: UserResourcePolicyProcessors
    project_resource_policy: ProjectResourcePolicyProcessors
    resource_preset: ResourcePresetProcessors
    utilization_metric: UtilizationMetricProcessors
    model_serving: ModelServingProcessors
    model_serving_auto_scaling: ModelServingAutoScalingProcessors
    auth: AuthProcessors
    object_storage: ObjectStorageProcessors
    artifact: ArtifactProcessors
    artifact_registry: ArtifactRegistryProcessors
    artifact_revision: ArtifactRevisionProcessors
    deployment: Optional[DeploymentProcessors]
    storage_namespace: StorageNamespaceProcessors

    @classmethod
    def create(cls, args: ProcessorArgs, action_monitors: list[ActionMonitor]) -> Self:
        services = Services.create(args.service_args)
        agent_processors = AgentProcessors(services.agent, action_monitors)
        domain_processors = DomainProcessors(services.domain, action_monitors)
        group_processors = GroupProcessors(services.group, action_monitors)
        user_processors = UserProcessors(services.user, action_monitors)
        image_processors = ImageProcessors(services.image, action_monitors)
        container_registry_processors = ContainerRegistryProcessors(
            services.container_registry, action_monitors
        )
        vfolder_processors = VFolderProcessors(services.vfolder, action_monitors)
        vfolder_file_processors = VFolderFileProcessors(services.vfolder_file, action_monitors)
        vfolder_invite_processors = VFolderInviteProcessors(
            services.vfolder_invite, action_monitors
        )
        session_processors = SessionProcessors(services.session, action_monitors)
        keypair_resource_policy_processors = KeypairResourcePolicyProcessors(
            services.keypair_resource_policy, action_monitors
        )
        user_resource_policy_processors = UserResourcePolicyProcessors(
            services.user_resource_policy, action_monitors
        )
        project_resource_policy_processors = ProjectResourcePolicyProcessors(
            services.project_resource_policy, action_monitors
        )
        resource_preset_processors = ResourcePresetProcessors(
            services.resource_preset, action_monitors
        )
        model_serving_processors = ModelServingProcessors(services.model_serving, action_monitors)
        model_serving_auto_scaling_processors = ModelServingAutoScalingProcessors(
            services.model_serving_auto_scaling, action_monitors
        )
        utilization_metric_processors = UtilizationMetricProcessors(
            services.utilization_metric, action_monitors
        )
        auth = AuthProcessors(services.auth, action_monitors)
        object_storage_processors = ObjectStorageProcessors(
            services.object_storage, action_monitors
        )
        artifact_processors = ArtifactProcessors(services.artifact, action_monitors)
        artifact_registry_processors = ArtifactRegistryProcessors(
            services.artifact_registry, action_monitors
        )
        artifact_revision_processors = ArtifactRevisionProcessors(
            services.artifact_revision, action_monitors
        )

        # Initialize deployment processors if service is available
        deployment_processors = None
        if services.deployment is not None:
            deployment_processors = DeploymentProcessors(services.deployment, action_monitors)

        storage_namespace_processors = StorageNamespaceProcessors(
            services.storage_namespace, action_monitors
        )

        return cls(
            agent=agent_processors,
            domain=domain_processors,
            group=group_processors,
            user=user_processors,
            image=image_processors,
            container_registry=container_registry_processors,
            vfolder=vfolder_processors,
            vfolder_file=vfolder_file_processors,
            vfolder_invite=vfolder_invite_processors,
            session=session_processors,
            keypair_resource_policy=keypair_resource_policy_processors,
            user_resource_policy=user_resource_policy_processors,
            project_resource_policy=project_resource_policy_processors,
            resource_preset=resource_preset_processors,
            utilization_metric=utilization_metric_processors,
            model_serving=model_serving_processors,
            model_serving_auto_scaling=model_serving_auto_scaling_processors,
            auth=auth,
            object_storage=object_storage_processors,
            artifact=artifact_processors,
            artifact_registry=artifact_registry_processors,
            artifact_revision=artifact_revision_processors,
            deployment=deployment_processors,
            storage_namespace=storage_namespace_processors,
        )

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            *self.agent.supported_actions(),
            *self.domain.supported_actions(),
            *self.group.supported_actions(),
            *self.user.supported_actions(),
            *self.image.supported_actions(),
            *self.container_registry.supported_actions(),
            *self.vfolder.supported_actions(),
            *self.vfolder_file.supported_actions(),
            *self.vfolder_invite.supported_actions(),
            *self.session.supported_actions(),
            *self.keypair_resource_policy.supported_actions(),
            *self.user_resource_policy.supported_actions(),
            *self.project_resource_policy.supported_actions(),
            *self.resource_preset.supported_actions(),
            *self.utilization_metric.supported_actions(),
            *self.model_serving.supported_actions(),
            *self.model_serving_auto_scaling.supported_actions(),
            *self.auth.supported_actions(),
            *self.object_storage.supported_actions(),
            *self.artifact_registry.supported_actions(),
            *self.artifact_revision.supported_actions(),
            *self.artifact.supported_actions(),
            *(self.deployment.supported_actions() if self.deployment else []),
            *self.storage_namespace.supported_actions(),
        ]



"""update_session_and_kernel_status

Revision ID: 6e44ea67d26e
Revises: d4c174934fd0
Create Date: 2024-08-05 14:39:11.901615

"""

import enum

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "6e44ea67d26e"
down_revision = "d4c174934fd0"
branch_labels = None
depends_on = None

KERNEL_STATUS_ENUM_NAME = "kernelstatus"
SESSION_STATUS_ENUM_NAME = "sessionstatus"

PREPARED = "PREPARED"


class OldKernelStatus(enum.Enum):
    PENDING = 0
    SCHEDULED = 5
    PREPARING = 10
    BUILDING = 20
    PULLING = 21
    RUNNING = 30
    RESTARTING = 31
    RESIZING = 32
    SUSPENDED = 33
    TERMINATING = 40
    TERMINATED = 41
    ERROR = 42
    CANCELLED = 43


class OldSessionStatus(enum.Enum):
    PENDING = 0
    SCHEDULED = 5
    PULLING = 7
    PREPARING = 10
    RUNNING = 30
    RESTARTING = 31
    RUNNING_DEGRADED = 32
    TERMINATING = 40
    TERMINATED = 41
    ERROR = 42
    CANCELLED = 43


def upgrade() -> None:
    conn = op.get_bind()
    # Drop indexes temporarily
    op.drop_index(op.f("ix_sessions_status"), table_name="sessions")
    op.drop_index(op.f("ix_kernels_status"), table_name="kernels")
    op.drop_index(op.f("ix_kernels_status_role"), table_name="kernels")
    op.drop_index(op.f("ix_kernels_unique_sess_token"), table_name="kernels")

    # Relax the colum type from enum to varchar(64).
    conn.execute(
        text("ALTER TABLE kernels ALTER COLUMN status TYPE VARCHAR(64) USING status::VARCHAR;")
    )
    conn.execute(text("ALTER TABLE kernels ALTER COLUMN status SET DEFAULT 'PENDING';"))
    conn.execute(
        text("ALTER TABLE sessions ALTER COLUMN status TYPE VARCHAR(64) USING status::VARCHAR;")
    )
    conn.execute(text("ALTER TABLE sessions ALTER COLUMN status SET DEFAULT 'PENDING';"))
    conn.execute(text(f"DROP TYPE {KERNEL_STATUS_ENUM_NAME};"))
    conn.execute(text(f"DROP TYPE {SESSION_STATUS_ENUM_NAME};"))

    # Recreate indexes
    op.create_index(op.f("ix_sessions_status"), "sessions", ["status"], unique=False)
    op.create_index(op.f("ix_kernels_status"), "kernels", ["status"], unique=False)


def downgrade() -> None:
    conn = op.get_bind()

    def _remove_new_state_and_create_enum(
        table_name: str, enum_name: str, status_enum: type[enum.Enum]
    ) -> None:
        conn.execute(
            text(
                f"""\
            UPDATE {table_name}
            SET status = CASE
                WHEN status = '{PREPARED}' THEN 'SCHEDULED'
                ELSE status
            END;
        """
            )
        )
        _values = ",".join(f"'{choice.name}'" for choice in status_enum)
        conn.execute(text(f"CREATE TYPE {enum_name} AS ENUM ({_values})"))
        conn.execute(text(f"ALTER TABLE {table_name} ALTER COLUMN status DROP DEFAULT;"))
        conn.execute(
            text(
                f"ALTER TABLE {table_name} ALTER COLUMN status TYPE {enum_name} USING status::{enum_name};"
            )
        )
        conn.execute(text(f"ALTER TABLE {table_name} ALTER COLUMN status SET DEFAULT 'PENDING';"))

    _remove_new_state_and_create_enum("kernels", KERNEL_STATUS_ENUM_NAME, OldKernelStatus)
    _remove_new_state_and_create_enum("sessions", SESSION_STATUS_ENUM_NAME, OldSessionStatus)

    op.create_index("ix_kernels_status_role", "kernels", ["status", "cluster_role"], unique=False)
    op.create_index(
        "ix_kernels_unique_sess_token",
        "kernels",
        ["access_key", "session_name"],
        unique=True,
        postgresql_where=sa.text(
            "status NOT IN ('TERMINATED', 'CANCELLED') and cluster_role = 'main'"
        ),
    )



from __future__ import annotations

import logging
from datetime import datetime

from ai.backend.common.clients.valkey_client.valkey_image.client import ValkeyImageClient
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.types import AgentId
from ai.backend.logging.utils import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AgentCacheSource:
    """
    Cache source for agent-related operations.
    Uses simple JSON serialization with short TTL for caching.
    """

    _valkey_image: ValkeyImageClient
    _valkey_live: ValkeyLiveClient

    def __init__(self, valkey_image: ValkeyImageClient, valkey_live: ValkeyLiveClient) -> None:
        self._valkey_image = valkey_image
        self._valkey_live = valkey_live

    async def set_agent_to_images(self, agent_id: AgentId, image_canonicals: list[str]) -> None:
        await self._valkey_image.add_agent_to_images(agent_id, image_canonicals)

    async def update_agent_last_seen(self, agent_id: AgentId, time: datetime) -> None:
        await self._valkey_live.update_agent_last_seen(agent_id, time.timestamp())

    async def remove_agent_last_seen(self, agent_id: AgentId) -> None:
        await self._valkey_live.remove_agent_last_seen(agent_id)

    async def remove_agent_from_all_images(self, agent_id: AgentId) -> None:
        await self._valkey_image.remove_agent_from_all_images(agent_id)

    async def remove_agent_from_images(
        self, agent_id: AgentId, image_canonicals: list[str]
    ) -> None:
        await self._valkey_image.remove_agent_from_images(agent_id, image_canonicals)



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.resource.types import KeyPairResourcePolicyData
from ai.backend.manager.services.keypair_resource_policy.actions.base import (
    KeypairResourcePolicyAction,
)


@dataclass
class DeleteKeyPairResourcePolicyAction(KeypairResourcePolicyAction):
    name: str

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "delete"


@dataclass
class DeleteKeyPairResourcePolicyActionResult(BaseActionResult):
    keypair_resource_policy: KeyPairResourcePolicyData

    @override
    def entity_id(self) -> Optional[str]:
        return self.keypair_resource_policy.name



import strawberry
from strawberry.federation import Schema
from strawberry.schema.config import StrawberryConfig

from .agent_stats import (
    agent_stats,
)
from .artifact import (
    approve_artifact_revision,
    artifact,
    artifact_import_progress_updated,
    artifact_revision,
    artifact_revisions,
    artifact_status_changed,
    artifacts,
    cancel_import_artifact,
    cleanup_artifact_revisions,
    delegate_import_artifacts,
    delegate_scan_artifacts,
    delete_artifacts,
    import_artifacts,
    reject_artifact_revision,
    restore_artifacts,
    scan_artifact_models,
    scan_artifacts,
    update_artifact,
)
from .artifact_registry import default_artifact_registry
from .huggingface_registry import (
    create_huggingface_registry,
    delete_huggingface_registry,
    huggingface_registries,
    huggingface_registry,
    update_huggingface_registry,
)
from .model_deployment.model_deployment import (
    create_model_deployment,
    delete_model_deployment,
    deployment,
    deployment_status_changed,
    deployments,
    replica,
    replica_status_changed,
    update_model_deployment,
)
from .model_deployment.model_revision import (
    create_model_revision,
    revision,
    revisions,
)
from .object_storage import (
    create_object_storage,
    delete_object_storage,
    get_presigned_download_url,
    get_presigned_upload_url,
    object_storage,
    object_storages,
    update_object_storage,
)
from .reservoir_registry import (
    create_reservoir_registry,
    delete_reservoir_registry,
    reservoir_registries,
    reservoir_registry,
    update_reservoir_registry,
)
from .storage_namespace import (
    register_storage_namespace,
    unregister_storage_namespace,
)


@strawberry.type
class Query:
    artifact = artifact
    artifacts = artifacts
    artifact_revision = artifact_revision
    artifact_revisions = artifact_revisions
    deployments = deployments
    deployment = deployment
    revisions = revisions
    revision = revision
    replica = replica
    object_storage = object_storage
    object_storages = object_storages
    huggingface_registry = huggingface_registry
    huggingface_registries = huggingface_registries
    reservoir_registry = reservoir_registry
    reservoir_registries = reservoir_registries
    default_artifact_registry = default_artifact_registry
    agent_stats = agent_stats


@strawberry.type
class Mutation:
    scan_artifacts = scan_artifacts
    scan_artifact_models = scan_artifact_models
    import_artifacts = import_artifacts
    delegate_scan_artifacts = delegate_scan_artifacts
    delegate_import_artifacts = delegate_import_artifacts
    update_artifact = update_artifact
    delete_artifacts = delete_artifacts
    restore_artifacts = restore_artifacts
    cleanup_artifact_revisions = cleanup_artifact_revisions
    cancel_import_artifact = cancel_import_artifact
    create_model_deployment = create_model_deployment
    update_model_deployment = update_model_deployment
    delete_model_deployment = delete_model_deployment
    create_model_revision = create_model_revision
    create_object_storage = create_object_storage
    update_object_storage = update_object_storage
    delete_object_storage = delete_object_storage
    register_storage_namespace = register_storage_namespace
    unregister_storage_namespace = unregister_storage_namespace
    create_huggingface_registry = create_huggingface_registry
    update_huggingface_registry = update_huggingface_registry
    delete_huggingface_registry = delete_huggingface_registry
    create_reservoir_registry = create_reservoir_registry
    update_reservoir_registry = update_reservoir_registry
    delete_reservoir_registry = delete_reservoir_registry
    get_presigned_download_url = get_presigned_download_url
    get_presigned_upload_url = get_presigned_upload_url
    approve_artifact_revision = approve_artifact_revision
    reject_artifact_revision = reject_artifact_revision


@strawberry.type
class Subscription:
    artifact_status_changed = artifact_status_changed
    artifact_import_progress_updated = artifact_import_progress_updated
    deployment_status_changed = deployment_status_changed
    replica_status_changed = replica_status_changed


class CustomizedSchema(Schema):
    def as_str(self) -> str:
        sdl = super().as_str()
        sdl = sdl.replace("type PageInfo", "type PageInfo @shareable").replace(
            'import: ["@external", "@key"]', 'import: ["@external", "@key", "@shareable"]'
        )
        # Convert escaped newlines to actual newlines for better description formatting
        sdl = sdl.replace("\\n", "\n")

        return sdl


schema = CustomizedSchema(
    query=Query,
    mutation=Mutation,
    subscription=Subscription,
    config=StrawberryConfig(auto_camel_case=True),
    enable_federation_2=True,
)



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec

from ..actions.invite import (
    AcceptInvitationAction,
    AcceptInvitationActionResult,
    InviteVFolderAction,
    InviteVFolderActionResult,
    LeaveInvitedVFolderAction,
    LeaveInvitedVFolderActionResult,
    ListInvitationAction,
    ListInvitationActionResult,
    RejectInvitationAction,
    RejectInvitationActionResult,
    RevokeInvitedVFolderAction,
    RevokeInvitedVFolderActionResult,
    UpdateInvitationAction,
    UpdateInvitationActionResult,
    UpdateInvitedVFolderMountPermissionAction,
    UpdateInvitedVFolderMountPermissionActionResult,
)
from ..services.invite import VFolderInviteService


class VFolderInviteProcessors(AbstractProcessorPackage):
    invite_vfolder: ActionProcessor[InviteVFolderAction, InviteVFolderActionResult]
    accept_invitation: ActionProcessor[AcceptInvitationAction, AcceptInvitationActionResult]
    reject_invitation: ActionProcessor[RejectInvitationAction, RejectInvitationActionResult]
    update_invitation: ActionProcessor[UpdateInvitationAction, UpdateInvitationActionResult]
    list_invitation: ActionProcessor[ListInvitationAction, ListInvitationActionResult]
    leave_invited_vfolder: ActionProcessor[
        LeaveInvitedVFolderAction, LeaveInvitedVFolderActionResult
    ]
    revoke_invited_vfolder: ActionProcessor[
        RevokeInvitedVFolderAction, RevokeInvitedVFolderActionResult
    ]
    update_invited_vfolder_mount_permission: ActionProcessor[
        UpdateInvitedVFolderMountPermissionAction, UpdateInvitedVFolderMountPermissionActionResult
    ]

    def __init__(self, service: VFolderInviteService, action_monitors: list[ActionMonitor]):
        self.invite_vfolder = ActionProcessor(service.invite, action_monitors)
        self.accept_invitation = ActionProcessor(service.accept_invitation, action_monitors)
        self.reject_invitation = ActionProcessor(service.reject_invitation, action_monitors)
        self.update_invitation = ActionProcessor(service.update_invitation, action_monitors)
        self.list_invitation = ActionProcessor(service.list_invitation, action_monitors)
        self.leave_invited_vfolder = ActionProcessor(service.leave_invited_vfolder, action_monitors)
        self.revoke_invited_vfolder = ActionProcessor(
            service.revoke_invited_vfolder, action_monitors
        )
        self.update_invited_vfolder_mount_permission = ActionProcessor(
            service.update_invited_vfolder_mount_permission, action_monitors
        )

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            InviteVFolderAction.spec(),
            AcceptInvitationAction.spec(),
            RejectInvitationAction.spec(),
            UpdateInvitationAction.spec(),
            ListInvitationAction.spec(),
            LeaveInvitedVFolderAction.spec(),
            RevokeInvitedVFolderAction.spec(),
            UpdateInvitedVFolderMountPermissionAction.spec(),
        ]



from dataclasses import dataclass, field
from typing import Any, Optional, override

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.image.types import ImageData
from ai.backend.manager.models.image import ImageType
from ai.backend.manager.services.image.actions.base import ImageAction
from ai.backend.manager.types import OptionalState, PartialModifier, TriState


@dataclass
class ImageModifier(PartialModifier):
    name: OptionalState[str] = field(default_factory=OptionalState.nop)
    registry: OptionalState[str] = field(default_factory=OptionalState.nop)
    image: OptionalState[str] = field(default_factory=OptionalState.nop)
    tag: OptionalState[str] = field(default_factory=OptionalState.nop)
    architecture: OptionalState[str] = field(default_factory=OptionalState.nop)
    is_local: OptionalState[bool] = field(default_factory=OptionalState.nop)
    size_bytes: OptionalState[int] = field(default_factory=OptionalState.nop)
    type: OptionalState[ImageType] = field(default_factory=OptionalState.nop)
    config_digest: OptionalState[str] = field(default_factory=OptionalState.nop)
    labels: OptionalState[dict[str, Any]] = field(default_factory=OptionalState.nop)
    accelerators: TriState[str] = field(default_factory=TriState.nop)
    resources: OptionalState[dict[str, Any]] = field(default_factory=OptionalState.nop)

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.registry.update_dict(to_update, "registry")
        self.image.update_dict(to_update, "image")
        self.tag.update_dict(to_update, "tag")
        self.architecture.update_dict(to_update, "architecture")
        self.is_local.update_dict(to_update, "is_local")
        self.size_bytes.update_dict(to_update, "size_bytes")
        self.type.update_dict(to_update, "type")
        self.config_digest.update_dict(to_update, "config_digest")
        self.labels.update_dict(to_update, "labels")
        self.accelerators.update_dict(to_update, "accelerators")
        self.resources.update_dict(to_update, "_resources")
        return to_update


@dataclass
class ModifyImageAction(ImageAction):
    target: str
    architecture: str
    modifier: ImageModifier

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "modify"


@dataclass
class ModifyImageActionResult(BaseActionResult):
    image: ImageData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.image.id)


class ModifyImageActionUnknownImageReferenceError(BackendAIError):
    error_type = "https://api.backend.ai/probs/image-not-found"
    error_title = "Unknown image reference."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.IMAGE,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.NOT_FOUND,
        )



import uuid

import aiotools
import sqlalchemy as sa

from ai.backend.common.types import VFolderID
from ai.backend.manager.errors.resource import (
    ProjectHasActiveEndpointsError,
    ProjectHasActiveKernelsError,
    ProjectHasVFoldersMountedError,
    ProjectNotFound,
)
from ai.backend.manager.errors.storage import VFolderOperationFailed
from ai.backend.manager.models.endpoint import EndpointLifecycle, EndpointRow
from ai.backend.manager.models.group import groups
from ai.backend.manager.models.routing import RoutingRow
from ai.backend.manager.models.session import SessionRow
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, SASession


class AdminGroupRepository:
    _db: ExtendedAsyncSAEngine
    _storage_manager: StorageSessionManager

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        storage_manager: StorageSessionManager,
    ) -> None:
        self._db = db
        self._storage_manager = storage_manager

    async def _check_group_vfolders_mounted_to_active_kernels(
        self, session: SASession, group_id: uuid.UUID
    ) -> bool:
        """Check if group has vfolders mounted to active kernels."""
        from ai.backend.manager.models import (
            AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES,
            kernels,
            vfolders,
        )

        # Get group vfolder IDs
        query = sa.select([vfolders.c.id]).select_from(vfolders).where(vfolders.c.group == group_id)
        result = await session.execute(query)
        rows = result.fetchall()
        group_vfolder_ids = [row["id"] for row in rows]

        # Check if any active kernels have these vfolders mounted
        query = (
            sa.select([kernels.c.mounts])
            .select_from(kernels)
            .where(
                (kernels.c.group_id == group_id)
                & (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES)),
            )
        )
        async for row in await session.stream(query):
            for _mount in row["mounts"]:
                try:
                    vfolder_id = uuid.UUID(_mount[2])
                    if vfolder_id in group_vfolder_ids:
                        return True
                except Exception:
                    pass
        return False

    async def _check_group_has_active_kernels(
        self, session: SASession, group_id: uuid.UUID
    ) -> bool:
        """Check if group has active kernels."""
        from ai.backend.manager.models import AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES, kernels

        query = (
            sa.select([sa.func.count()])
            .select_from(kernels)
            .where(
                (kernels.c.group_id == group_id)
                & (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES))
            )
        )
        active_kernel_count = await session.scalar(query)
        return True if active_kernel_count > 0 else False

    async def _delete_group_vfolders(self, group_id: uuid.UUID) -> int:
        """Delete all vfolders belonging to the group."""
        from typing import cast

        from ai.backend.manager.models import (
            VFolderDeletionInfo,
            VFolderRow,
            VFolderStatusSet,
            initiate_vfolder_deletion,
            vfolder_status_map,
        )

        target_vfs: list[VFolderDeletionInfo] = []
        async with self._db.begin_session() as session:
            query = sa.select(VFolderRow).where(
                sa.and_(
                    VFolderRow.group == group_id,
                    VFolderRow.status.in_(vfolder_status_map[VFolderStatusSet.DELETABLE]),
                )
            )
            result = await session.scalars(query)
            rows = cast(list[VFolderRow], result.fetchall())
            for vf in rows:
                target_vfs.append(
                    VFolderDeletionInfo(VFolderID.from_row(vf), vf.host, vf.unmanaged_path)
                )

        storage_ptask_group = aiotools.PersistentTaskGroup()
        try:
            await initiate_vfolder_deletion(
                self._db,
                target_vfs,
                self._storage_manager,
                storage_ptask_group,
            )
        except VFolderOperationFailed:
            raise

        return len(target_vfs)

    async def _delete_group_kernels(self, session: SASession, group_id: uuid.UUID) -> int:
        """Delete all kernels belonging to the group."""
        from ai.backend.manager.models import kernels

        query = sa.delete(kernels).where(kernels.c.group_id == group_id)
        result = await session.execute(query)
        return result.rowcount

    async def _delete_group_sessions(self, session: SASession, group_id: uuid.UUID) -> int:
        """Delete all sessions belonging to the group."""
        stmt = sa.delete(SessionRow).where(SessionRow.group_id == group_id)
        result = await session.execute(stmt)
        return result.rowcount

    async def _delete_group_endpoints(self, session: SASession, group_id: uuid.UUID) -> None:
        """Delete all endpoints belonging to the group."""
        # Get all endpoints for the group
        endpoints = (
            await session.execute(
                sa.select(
                    EndpointRow.id,
                    sa.case(
                        (
                            EndpointRow.lifecycle_stage.in_([
                                EndpointLifecycle.CREATED,
                                EndpointLifecycle.DESTROYING,
                            ]),
                            True,
                        ),
                        else_=False,
                    ).label("is_active"),
                ).where(EndpointRow.project == group_id)
            )
        ).all()

        if len(endpoints) == 0:
            return

        # Check for active endpoints
        active_endpoints = [ep.id for ep in endpoints if ep.is_active]
        if len(active_endpoints) > 0:
            raise ProjectHasActiveEndpointsError(f"project {group_id} has active endpoints")

        # Delete endpoint-related data
        endpoint_ids = [ep.id for ep in endpoints]

        # Get session IDs before deleting endpoints
        session_ids_result = await session.scalars(
            sa.select(RoutingRow.session).where(
                sa.and_(
                    RoutingRow.endpoint.in_(endpoint_ids),
                    RoutingRow.session.is_not(None),
                )
            )
        )
        session_ids = list(session_ids_result.unique().all())

        # Delete endpoints (routings are CASCADE deleted automatically)
        await session.execute(
            sa.delete(EndpointRow).where(EndpointRow.id.in_(endpoint_ids)),
            execution_options={"synchronize_session": False},
        )

        # Delete sessions that were associated with endpoints
        if session_ids:
            await session.execute(
                sa.delete(SessionRow).where(SessionRow.id.in_(session_ids)),
                execution_options={"synchronize_session": False},
            )

    async def purge_group_force(self, group_id: uuid.UUID) -> bool:
        """Completely remove a group and all its associated data."""
        async with self._db.begin_session() as session:
            # Pre-flight checks
            if await self._check_group_vfolders_mounted_to_active_kernels(session, group_id):
                raise ProjectHasVFoldersMountedError(
                    f"error on deleting project {group_id} with vfolders mounted to active kernels"
                )

            if await self._check_group_has_active_kernels(session, group_id):
                raise ProjectHasActiveKernelsError(
                    f"error on deleting project {group_id} with active kernels"
                )

            # Delete associated resources
            await self._delete_group_endpoints(session, group_id)

            # Commit session before vfolder deletion (which uses separate transactions)
            await session.commit()

        # Delete vfolders (uses separate transaction)
        await self._delete_group_vfolders(group_id)

        async with self._db.begin_session() as session:
            # Delete remaining data
            await self._delete_group_kernels(session, group_id)
            await self._delete_group_sessions(session, group_id)

            # Finally delete the group itself
            result = await session.execute(sa.delete(groups).where(groups.c.id == group_id))

            if result.rowcount > 0:
                return True
            raise ProjectNotFound("project not found")



import uuid
from typing import Optional, cast

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import contains_eager

from ...data.permission.id import ObjectId
from ...data.permission.role import (
    RoleCreateInput,
    RoleDeleteInput,
    RoleUpdateInput,
    UserRoleAssignmentInput,
)
from ...data.permission.status import (
    RoleStatus,
)
from ...errors.common import ObjectNotFound
from ...models.rbac_models.association_scopes_entities import AssociationScopesEntitiesRow
from ...models.rbac_models.permission.object_permission import ObjectPermissionRow
from ...models.rbac_models.permission.permission import PermissionRow
from ...models.rbac_models.permission.permission_group import PermissionGroupRow
from ...models.rbac_models.role import RoleRow
from ...models.rbac_models.user_role import UserRoleRow
from ...models.utils import ExtendedAsyncSAEngine


class PermissionDBSource:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def create_role(self, data: RoleCreateInput) -> RoleRow:
        """
        Create a new role in the database.

        Returns the ID of the created role.
        """
        async with self._db.begin_session() as db_session:
            role_row = RoleRow.from_input(data)
            db_session.add(role_row)  # type: ignore[arg-type]
            await db_session.flush()
            role_id = role_row.id
            for permission_group in data.permission_groups:
                permission_group_row = PermissionGroupRow.from_input(
                    permission_group.to_input(role_id)
                )
                db_session.add(permission_group_row)  # type: ignore[arg-type]
            for object_permission in data.object_permissions:
                object_permission_row = ObjectPermissionRow.from_input(
                    object_permission.to_input(role_id)
                )
                db_session.add(object_permission_row)  # type: ignore[arg-type]
            await db_session.flush()
            await db_session.refresh(role_row)
            return role_row

    async def _get_role(self, db_session: SASession, role_id: uuid.UUID) -> RoleRow:
        stmt = sa.select(RoleRow).where(RoleRow.id == role_id)
        role_row = await db_session.scalar(stmt)
        result = cast(Optional[RoleRow], role_row)
        if result is None:
            raise ObjectNotFound(f"Role with ID {role_id} does not exist.")
        return result

    async def update_role(self, data: RoleUpdateInput) -> RoleRow:
        to_update = data.fields_to_update()
        async with self._db.begin_session() as db_session:
            stmt = sa.update(RoleRow).where(RoleRow.id == data.id).values(**to_update)
            await db_session.execute(stmt)
            role_row = await self._get_role(db_session, data.id)
            return role_row

    async def delete_role(self, data: RoleDeleteInput) -> RoleRow:
        async with self._db.begin_session() as db_session:
            role_row = await self._get_role(db_session, data.id)
            role_row.status = RoleStatus.DELETED
            await db_session.flush()
            await db_session.refresh(role_row)
            return role_row

    async def assign_role(self, data: UserRoleAssignmentInput) -> UserRoleRow:
        async with self._db.begin_session() as db_session:
            user_role_row = UserRoleRow.from_input(data)
            db_session.add(user_role_row)  # type: ignore[arg-type]
            await db_session.flush()
            await db_session.refresh(user_role_row)
            return user_role_row

    async def get_role(self, role_id: uuid.UUID) -> Optional[RoleRow]:
        async with self._db.begin_readonly_session() as db_session:
            try:
                result = await self._get_role(db_session, role_id)
            except ObjectNotFound:
                return None
            return result

    async def get_user_roles(self, user_id: uuid.UUID) -> list[RoleRow]:
        async with self._db.begin_readonly_session() as db_session:
            j = (
                sa.join(
                    RoleRow,
                    UserRoleRow,
                    RoleRow.id == UserRoleRow.role_id,
                )
                .join(
                    ObjectPermissionRow,
                    RoleRow.id == ObjectPermissionRow.role_id,
                )
                .join(
                    PermissionGroupRow,
                    RoleRow.id == PermissionGroupRow.role_id,
                )
                .join(
                    PermissionRow,
                    PermissionGroupRow.id == PermissionRow.permission_group_id,
                )
            )
            stmt = (
                sa.select(RoleRow)
                .select_from(j)
                .where(UserRoleRow.user_id == user_id)
                .options(
                    contains_eager(RoleRow.permission_group_rows).options(
                        contains_eager(PermissionGroupRow.permission_rows)
                    ),
                    contains_eager(RoleRow.object_permission_rows),
                )
            )

            result = await db_session.scalars(stmt)
            return result.all()

    async def get_entity_mapped_scopes(
        self, target_object_id: ObjectId
    ) -> list[AssociationScopesEntitiesRow]:
        async with self._db.begin_readonly_session() as db_session:
            stmt = sa.select(AssociationScopesEntitiesRow.scope_id).where(
                AssociationScopesEntitiesRow.entity_id == target_object_id.entity_id,
                AssociationScopesEntitiesRow.entity_type == target_object_id.entity_type.value,
            )
            result = await db_session.scalars(stmt)
            return result.all()



from dataclasses import dataclass
from typing import Any, Mapping, Optional, override

from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.session.types import SessionData
from ai.backend.manager.services.session.base import SessionAction


@dataclass
class CommitSessionAction(SessionAction):
    session_name: str
    owner_access_key: AccessKey
    filename: Optional[str]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "commit"


@dataclass
class CommitSessionActionResult(BaseActionResult):
    session_data: SessionData

    # TODO: Add proper type
    commit_result: Mapping[str, Any]

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.session_data.id)



import asyncio
import logging
from dataclasses import dataclass
from typing import override

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.reporters.base import (
    AbstractReporter,
    FinishedActionMessage,
    StartedActionMessage,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@dataclass
class ReporterHubArgs:
    reporters: dict[str, list[AbstractReporter]]  # Key: action type, Value: reporter instance


class ReporterHub(AbstractReporter):
    _start_queue: asyncio.Queue[StartedActionMessage]
    _finish_queue: asyncio.Queue[FinishedActionMessage]
    _reporters: dict[str, list[AbstractReporter]]  # Key: action type, Value: reporters list
    _closed: bool
    _start_task: asyncio.Task
    _finish_task: asyncio.Task

    def __init__(self, args: ReporterHubArgs) -> None:
        self._start_queue = asyncio.Queue()
        self._finish_queue = asyncio.Queue()
        self._reporters = args.reporters
        self._closed = False
        self._start_task = asyncio.create_task(self._report_started())
        self._finish_task = asyncio.create_task(self._report_finished())

    @override
    async def report_started(self, message: StartedActionMessage) -> None:
        await self._start_queue.put(message)

    @override
    async def report_finished(self, message: FinishedActionMessage) -> None:
        await self._finish_queue.put(message)

    def _target_reporters(self, action_type: str) -> list[AbstractReporter]:
        return self._reporters.get(action_type, [])

    async def _report_started(self) -> None:
        while not self._closed:
            message = await self._start_queue.get()
            target_reporters = self._target_reporters(message.action_type)
            for reporter in target_reporters:
                try:
                    await reporter.report_started(message)
                except Exception as e:
                    log.error(f"reporter.report_started failed: {e}")

    async def _report_finished(self) -> None:
        while not self._closed:
            message = await self._finish_queue.get()
            target_reporters = self._target_reporters(message.action_type)
            for reporter in target_reporters:
                try:
                    await reporter.report_finished(message)
                except Exception as e:
                    log.error(f"reporter.report_finished failed: {e}")

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        self._start_task.cancel()
        self._finish_task.cancel()



"""Validator for pending session resource limits."""

from ai.backend.common.types import ResourceSlot
from ai.backend.manager.sokovan.scheduler.types import SessionWorkload, SystemSnapshot

from .exceptions import PendingSessionResourceLimitExceeded
from .validator import ValidatorRule


class PendingSessionResourceLimitValidator(ValidatorRule):
    """
    Check if creating a new pending session would exceed the allowed resource limits.
    This corresponds to check_pending_session_resource_limit predicate.
    """

    def name(self) -> str:
        """Return the validator name for predicates."""
        return "PendingSessionResourceLimitValidator"

    def success_message(self) -> str:
        """Return a message describing successful validation."""
        return "Pending session resource usage is within the allowed limit for the access key"

    def validate(self, snapshot: SystemSnapshot, workload: SessionWorkload) -> None:
        # Get the keypair's resource policy
        policy = snapshot.resource_policy.keypair_policies.get(workload.access_key)
        if not policy:
            # If no policy is defined, we can't validate - let it pass
            return

        # Check if there's a pending session resource limit
        pending_resource_limit = policy.max_pending_session_resource_slots
        if not pending_resource_limit:
            # No limit set, allow
            return

        # Calculate current pending session resource usage
        pending_sessions = snapshot.pending_sessions.by_keypair.get(workload.access_key, [])
        current_pending_slots = ResourceSlot()
        for session in pending_sessions:
            current_pending_slots += session.requested_slots

        # Check if adding this workload would exceed the limit
        total_after = current_pending_slots
        if total_after > pending_resource_limit:
            # Format the current usage for human-readable output
            usage_str = " ".join(f"{k}={v}" for k, v in current_pending_slots.items() if v)
            raise PendingSessionResourceLimitExceeded(
                f"Your pending session quota is exceeded. ({usage_str})"
            )



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.image.types import ImageData
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.services.image.actions.base import ImageAction


@dataclass
class ForgetImageByIdAction(ImageAction):
    user_id: uuid.UUID
    client_role: UserRole
    image_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.image_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "forget_by_id"


@dataclass
class ForgetImageByIdActionResult(BaseActionResult):
    image: ImageData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.image.id)



import enum
from collections.abc import Mapping
from dataclasses import dataclass, field
from datetime import datetime
from functools import lru_cache
from typing import Any, Optional
from uuid import UUID

import yarl

from ai.backend.common.types import ClusterMode, RuntimeVariant, SessionId, VFolderMount
from ai.backend.manager.data.deployment.scale import AutoScalingRule
from ai.backend.manager.data.image.types import ImageIdentifier


class EndpointLifecycle(enum.Enum):
    PENDING = "pending"
    CREATED = "created"  # Deprecated, use READY instead
    SCALING = "scaling"
    READY = "ready"
    DESTROYING = "destroying"
    DESTROYED = "destroyed"

    @classmethod
    @lru_cache(maxsize=1)
    def active_states(cls) -> set["EndpointLifecycle"]:
        return {cls.PENDING, cls.CREATED, cls.SCALING, cls.READY}

    @classmethod
    @lru_cache(maxsize=1)
    def need_scaling_states(cls) -> set["EndpointLifecycle"]:
        return {cls.CREATED, cls.READY}

    @classmethod
    @lru_cache(maxsize=1)
    def inactive_states(cls) -> set["EndpointLifecycle"]:
        return {cls.DESTROYING, cls.DESTROYED}


class RouteStatus(enum.Enum):
    PROVISIONING = "provisioning"
    HEALTHY = "healthy"
    UNHEALTHY = "unhealthy"
    TERMINATING = "terminating"
    TERMINATED = "terminated"
    FAILED_TO_START = "failed_to_start"

    @classmethod
    @lru_cache(maxsize=1)
    def active_route_statuses(cls) -> set["RouteStatus"]:
        return {RouteStatus.PROVISIONING, RouteStatus.HEALTHY, RouteStatus.UNHEALTHY}

    @classmethod
    @lru_cache(maxsize=1)
    def inactive_route_statuses(cls) -> set["RouteStatus"]:
        return {RouteStatus.TERMINATING, RouteStatus.TERMINATED, RouteStatus.FAILED_TO_START}

    def is_active(self) -> bool:
        return self in self.active_route_statuses()

    def is_inactive(self) -> bool:
        return self in self.inactive_route_statuses()

    def termination_priority(self) -> int:
        priority_map = {
            RouteStatus.UNHEALTHY: 1,
            RouteStatus.PROVISIONING: 2,
            RouteStatus.HEALTHY: 3,
        }
        return priority_map.get(self, 0)


@dataclass
class DeploymentMetadata:
    name: str
    domain: str
    project: UUID
    resource_group: str
    created_user: UUID
    session_owner: UUID
    created_at: Optional[datetime]
    tag: Optional[str] = None


@dataclass
class DeploymentState:
    lifecycle: EndpointLifecycle
    retry_count: int


@dataclass
class MountSpec:
    mounts: list[UUID]
    mount_map: Mapping[UUID, str]
    mount_options: Mapping[UUID, dict[str, Any]]


@dataclass
class MountMetadata:
    model_vfolder_id: UUID
    model_definition_path: Optional[str] = None
    model_mount_destination: str = "/models"
    extra_mounts: list[VFolderMount] = field(default_factory=list)

    def to_mount_spec(self) -> MountSpec:
        mounts = [
            self.model_vfolder_id,
            *[m.vfid.folder_id for m in self.extra_mounts],
        ]
        mount_map = {
            self.model_vfolder_id: self.model_mount_destination,
            **{m.vfid.folder_id: m.kernel_path.as_posix() for m in self.extra_mounts},
        }
        mount_options = {m.vfid.folder_id: {"permission": m.mount_perm} for m in self.extra_mounts}
        return MountSpec(mounts=mounts, mount_map=mount_map, mount_options=mount_options)


@dataclass
class ReplicaSpec:
    replica_count: int
    desired_replica_count: Optional[int] = None

    @property
    def target_replica_count(self) -> int:
        if self.desired_replica_count is not None:
            return self.desired_replica_count
        return self.replica_count


@dataclass
class ResourceSpec:
    cluster_mode: ClusterMode
    cluster_size: int
    resource_slots: Mapping[str, Any]
    resource_opts: Optional[Mapping[str, Any]] = None


@dataclass
class ExecutionSpec:
    startup_command: Optional[str] = None
    bootstrap_script: Optional[str] = None
    environ: Optional[dict[str, str]] = None
    runtime_variant: RuntimeVariant = RuntimeVariant.CUSTOM
    callback_url: Optional[yarl.URL] = None


@dataclass
class ModelRevisionSpec:
    image_identifier: ImageIdentifier
    resource_spec: ResourceSpec
    mounts: MountMetadata
    execution: ExecutionSpec


@dataclass
class DeploymentNetworkSpec:
    open_to_public: bool
    url: Optional[str] = None


@dataclass
class DeploymentInfo:
    id: UUID
    metadata: DeploymentMetadata
    state: DeploymentState
    replica_spec: ReplicaSpec
    network: DeploymentNetworkSpec
    model_revisions: list[ModelRevisionSpec]

    def target_revision(self) -> Optional[ModelRevisionSpec]:
        if self.model_revisions:
            return self.model_revisions[0]
        return None


@dataclass
class DeploymentSessionSpec:
    id: UUID
    metadata: DeploymentMetadata


@dataclass
class ScaleOutDecision:
    deployment_info: DeploymentInfo
    new_replica_count: int


@dataclass
class DefinitionFiles:
    service_definition: Optional[dict[str, Any]]
    model_definition: dict[str, Any]


@dataclass
class RouteInfo:
    """Route information for deployment."""

    route_id: UUID
    endpoint_id: UUID
    session_id: Optional[SessionId]
    status: RouteStatus
    traffic_ratio: float
    created_at: datetime
    error_data: dict[str, Any] = field(default_factory=dict)


@dataclass
class DeploymentInfoWithRoutes:
    """DeploymentInfo with its routes."""

    deployment_info: DeploymentInfo
    routes: list[RouteInfo] = field(default_factory=list)


@dataclass
class DeploymentInfoWithAutoScalingRules:
    """DeploymentInfo with its autoscaling rules."""

    deployment_info: DeploymentInfo
    rules: list[AutoScalingRule] = field(default_factory=list)



"""add_public_host_to_agent

Revision ID: fdc9d6ac49b4
Revises: fe59ec332c07
Create Date: 2023-04-25 17:36:01.666605

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "fdc9d6ac49b4"
down_revision = "fe59ec332c07"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("agents", sa.Column("public_host", sa.String(length=256), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("agents", "public_host")
    # ### end Alembic commands ###



"""create endpoint_auto_scaling_rules table

Revision ID: fb89f5d7817b
Revises: ef9a7960d234
Create Date: 2024-12-20 01:48:21.009056

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID, IDColumn

# revision identifiers, used by Alembic.
revision = "fb89f5d7817b"
down_revision = "ef9a7960d234"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "endpoint_auto_scaling_rules",
        IDColumn(),
        sa.Column("metric_source", sa.VARCHAR(64), nullable=False),
        sa.Column("metric_name", sa.Text(), nullable=False),
        sa.Column("threshold", sa.Text(), nullable=False),
        sa.Column("comparator", sa.VARCHAR(64), nullable=False),
        sa.Column("step_size", sa.Integer(), nullable=False),
        sa.Column("cooldown_seconds", sa.Integer(), nullable=False),
        sa.Column("min_replicas", sa.Integer(), nullable=True),
        sa.Column("max_replicas", sa.Integer(), nullable=True),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
        sa.Column("last_triggered_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("endpoint", GUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["endpoint"],
            ["endpoints.id"],
            name=op.f("fk_endpoint_auto_scaling_rules_endpoint_endpoints"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_endpoint_auto_scaling_rules")),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("endpoint_auto_scaling_rules")
    # ### end Alembic commands ###



from __future__ import annotations

import logging
from typing import Final, Iterable, Tuple

import attrs
from aiohttp import web
from aiotools import apartial

from ai.backend.common.clients.valkey_client.valkey_rate_limit.client import ValkeyRateLimitClient
from ai.backend.common.defs import REDIS_RATE_LIMIT_DB, RedisRole
from ai.backend.logging import BraceStyleAdapter

from ..errors.api import RateLimitExceeded
from .context import RootContext
from .types import CORSOptions, WebMiddleware, WebRequestHandler

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

_rlim_window: Final = 60 * 15

# We implement rate limiting using a rolling counter, which prevents
# last-minute and first-minute bursts between the intervals.


@web.middleware
async def rlim_middleware(
    app: web.Application,
    request: web.Request,
    handler: WebRequestHandler,
) -> web.StreamResponse:
    # This is a global middleware: request.app is the root app.
    app_ctx: PrivateContext = app["ratelimit.context"]
    if request["is_authorized"]:
        rate_limit = request["keypair"]["rate_limit"]
        access_key = request["keypair"]["access_key"]
        rolling_count = await app_ctx.valkey_rate_limit_client.execute_rate_limit_logic(
            access_key=access_key,
            window=_rlim_window,
        )
        if rate_limit is not None and rolling_count > rate_limit:
            raise RateLimitExceeded
        remaining = rate_limit - rolling_count if rate_limit is not None else rolling_count
        response = await handler(request)
        response.headers["X-RateLimit-Limit"] = str(rate_limit)
        response.headers["X-RateLimit-Remaining"] = str(remaining)
        response.headers["X-RateLimit-Window"] = str(_rlim_window)
        return response
    else:
        # No checks for rate limiting for non-authorized queries.
        response = await handler(request)
        response.headers["X-RateLimit-Limit"] = "1000"
        response.headers["X-RateLimit-Remaining"] = "1000"
        response.headers["X-RateLimit-Window"] = str(_rlim_window)
        return response


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    valkey_rate_limit_client: ValkeyRateLimitClient
    redis_rlim_script: str


async def init(app: web.Application) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["ratelimit.context"]
    valkey_profile_target = root_ctx.config_provider.config.redis.to_valkey_profile_target()
    valkey_target = valkey_profile_target.profile_target(RedisRole.RATE_LIMIT)
    app_ctx.valkey_rate_limit_client = await ValkeyRateLimitClient.create(
        valkey_target=valkey_target,
        db_id=REDIS_RATE_LIMIT_DB,
        human_readable_name="ratelimit",
    )
    # Note: Script functionality is now handled internally by the client
    app_ctx.redis_rlim_script = ""


async def shutdown(app: web.Application) -> None:
    app_ctx: PrivateContext = app["ratelimit.context"]
    await app_ctx.valkey_rate_limit_client.flush_database()
    await app_ctx.valkey_rate_limit_client.close()


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    # default_cors_options is kept for API consistency but not used in rate limiting
    app = web.Application()
    app["api_versions"] = (1, 2, 3, 4)
    app["ratelimit.context"] = PrivateContext()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    # middleware must be wrapped by web.middleware at the outermost level.
    return app, [web.middleware(apartial(rlim_middleware, app))]



import logging

from ai.backend.common.dto.manager.rpc_request import PurgeImagesReq
from ai.backend.common.exception import UnknownImageReference
from ai.backend.common.types import AgentId, ImageAlias
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.errors.image import ImageNotFound
from ai.backend.manager.models.image import (
    ImageIdentifier,
)
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.repositories.image.admin_repository import AdminImageRepository
from ai.backend.manager.repositories.image.repository import ImageRepository
from ai.backend.manager.services.image.actions.alias_image import (
    AliasImageAction,
    AliasImageActionResult,
)
from ai.backend.manager.services.image.actions.clear_image_custom_resource_limit import (
    ClearImageCustomResourceLimitAction,
    ClearImageCustomResourceLimitActionResult,
)
from ai.backend.manager.services.image.actions.dealias_image import (
    DealiasImageAction,
    DealiasImageActionResult,
)
from ai.backend.manager.services.image.actions.forget_image import (
    ForgetImageAction,
    ForgetImageActionResult,
)
from ai.backend.manager.services.image.actions.forget_image_by_id import (
    ForgetImageByIdAction,
    ForgetImageByIdActionResult,
)
from ai.backend.manager.services.image.actions.modify_image import (
    ModifyImageAction,
    ModifyImageActionResult,
    ModifyImageActionUnknownImageReferenceError,
)
from ai.backend.manager.services.image.actions.preload_image import (
    PreloadImageAction,
    PreloadImageActionResult,
)
from ai.backend.manager.services.image.actions.purge_image_by_id import (
    PurgeImageByIdAction,
    PurgeImageByIdActionResult,
)
from ai.backend.manager.services.image.actions.purge_images import (
    PurgedImagesData,
    PurgeImageAction,
    PurgeImageActionResult,
    PurgeImagesAction,
    PurgeImagesActionResult,
)
from ai.backend.manager.services.image.actions.scan_image import (
    ScanImageAction,
    ScanImageActionResult,
)
from ai.backend.manager.services.image.actions.unload_image import (
    UnloadImageAction,
    UnloadImageActionResult,
)
from ai.backend.manager.services.image.actions.untag_image_from_registry import (
    UntagImageFromRegistryAction,
    UntagImageFromRegistryActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore


class ImageService:
    _agent_registry: AgentRegistry
    _image_repository: ImageRepository
    _admin_image_repository: AdminImageRepository

    def __init__(
        self,
        agent_registry: AgentRegistry,
        image_repository: ImageRepository,
        admin_image_repository: AdminImageRepository,
    ) -> None:
        self._agent_registry = agent_registry
        self._image_repository = image_repository
        self._admin_image_repository = admin_image_repository

    async def forget_image(self, action: ForgetImageAction) -> ForgetImageActionResult:
        if action.client_role == UserRole.SUPERADMIN:
            # Superadmin can forget any image without checking ownership
            data = await self._admin_image_repository.soft_delete_image_force(
                [
                    ImageIdentifier(action.reference, action.architecture),
                    ImageAlias(action.reference),
                ],
            )
            return ForgetImageActionResult(image=data)
        # Regular users can only forget images they own
        data = await self._image_repository.soft_delete_user_image(
            [
                ImageIdentifier(action.reference, action.architecture),
                ImageAlias(action.reference),
            ],
            action.user_id,
        )
        return ForgetImageActionResult(image=data)

    async def forget_image_by_id(
        self, action: ForgetImageByIdAction
    ) -> ForgetImageByIdActionResult:
        if action.client_role == UserRole.SUPERADMIN:
            # Superadmin can forget any image without checking ownership
            data = await self._admin_image_repository.soft_delete_image_by_id_force(action.image_id)
            return ForgetImageByIdActionResult(image=data)

        # Regular users need ownership validation
        data = await self._image_repository.soft_delete_image_by_id(action.image_id, action.user_id)
        return ForgetImageByIdActionResult(image=data)

    async def alias_image(self, action: AliasImageAction) -> AliasImageActionResult:
        try:
            image_id, image_alias = await self._image_repository.add_image_alias(
                action.alias, action.image_canonical, action.architecture
            )
        except UnknownImageReference:
            raise ImageNotFound
        return AliasImageActionResult(
            image_id=image_id,
            image_alias=image_alias,
        )

    async def dealias_image(self, action: DealiasImageAction) -> DealiasImageActionResult:
        result = await self._image_repository.delete_image_alias(action.alias)
        image_id, alias_data = result
        return DealiasImageActionResult(
            image_id=image_id,
            image_alias=alias_data,
        )

    async def modify_image(self, action: ModifyImageAction) -> ModifyImageActionResult:
        props = action.modifier

        try:
            to_update = props.fields_to_update()
            image_data = await self._image_repository.update_image_properties(
                action.target, action.architecture, to_update
            )
        except UnknownImageReference:
            raise ModifyImageActionUnknownImageReferenceError

        return ModifyImageActionResult(image=image_data)

    async def preload_image(self, action: PreloadImageAction) -> PreloadImageActionResult:
        raise NotImplementedError

    async def unload_image(self, action: UnloadImageAction) -> UnloadImageActionResult:
        raise NotImplementedError

    async def purge_image_by_id(self, action: PurgeImageByIdAction) -> PurgeImageByIdActionResult:
        if action.client_role == UserRole.SUPERADMIN:
            # Superadmin can delete any image without checking ownership
            image_data = await self._admin_image_repository.delete_image_with_aliases_force(
                action.image_id
            )
            return PurgeImageByIdActionResult(image=image_data)

        # Regular users need ownership validation
        image_data = await self._image_repository.delete_image_with_aliases_validated(
            action.image_id, action.user_id
        )
        return PurgeImageByIdActionResult(image=image_data)

    async def untag_image_from_registry(
        self, action: UntagImageFromRegistryAction
    ) -> UntagImageFromRegistryActionResult:
        if action.client_role == UserRole.SUPERADMIN:
            # Superadmin can untag without ownership check
            image_data = await self._admin_image_repository.untag_image_from_registry_force(
                action.image_id
            )
            return UntagImageFromRegistryActionResult(image=image_data)

        # Regular users need ownership validation
        image_data = await self._image_repository.untag_image_from_registry_validated(
            action.image_id, action.user_id
        )
        return UntagImageFromRegistryActionResult(image=image_data)

    async def purge_image(self, action: PurgeImageAction) -> PurgeImageActionResult:
        force, noprune = action.force, action.noprune
        agent_id = action.agent_id
        image_canonical = action.image.name
        arch = action.image.architecture

        image_identifier = ImageIdentifier(image_canonical, arch)
        image_data = await self._image_repository.resolve_image([image_identifier])

        results = await self._agent_registry.purge_images(
            AgentId(agent_id),
            PurgeImagesReq(images=[image_canonical], force=force, noprune=noprune),
        )

        error = None
        for result in results.responses:
            if result.error:
                error = (
                    f"Failed to purge image {image_canonical} from agent {agent_id}: {result.error}"
                )

        return PurgeImageActionResult(
            purged_image=image_data,
            error=error,
            reserved_bytes=image_data.size_bytes,
        )

    # TODO: The query is incorrectly separated. Need to optimize this.
    async def purge_images(self, action: PurgeImagesAction) -> PurgeImagesActionResult:
        errors = []
        total_reserved_bytes = 0
        force, noprune = action.force, action.noprune

        purged_images_data_list: list[PurgedImagesData] = []
        for key in action.keys:
            agent_id = key.agent_id
            images = key.images

            image_canonicals = [image.name for image in images]
            arch_per_images = {image.name: image.architecture for image in images}

            results = await self._agent_registry.purge_images(
                AgentId(agent_id),
                PurgeImagesReq(images=image_canonicals, force=force, noprune=noprune),
            )

            # Collect successful purges for batch resolution
            successful_identifiers: list[list[ImageIdentifier]] = []
            successful_canonicals = []

            for result in results.responses:
                if not result.error:
                    image_canonical = result.image
                    arch = arch_per_images[image_canonical]
                    successful_identifiers.append([ImageIdentifier(image_canonical, arch)])
                    successful_canonicals.append(image_canonical)
                else:
                    errors.append(
                        f"Failed to purge image {result.image} from agent {agent_id}: {result.error}"
                    )

            # Batch resolve all successful images
            purged_images_data = PurgedImagesData(purged_images=[], agent_id=agent_id)
            if successful_identifiers:
                image_data_list = await self._image_repository.resolve_images_batch(
                    successful_identifiers
                )
                for image_data, canonical in zip(image_data_list, successful_canonicals):
                    purged_images_data.purged_images.append(canonical)
                    total_reserved_bytes += image_data.size_bytes

            purged_images_data_list.append(purged_images_data)

        return PurgeImagesActionResult(
            purged_images=purged_images_data_list,
            errors=errors,
            total_reserved_bytes=total_reserved_bytes,
        )

    async def scan_image(self, action: ScanImageAction) -> ScanImageActionResult:
        image_canonical = action.canonical
        architecture = action.architecture

        result = await self._image_repository.scan_image_by_identifier(
            image_canonical, architecture
        )
        return ScanImageActionResult(image=result.images[0], errors=result.errors)

    async def clear_image_custom_resource_limit(
        self, action: ClearImageCustomResourceLimitAction
    ) -> ClearImageCustomResourceLimitActionResult:
        image_data = await self._image_repository.clear_image_custom_resource_limit(
            action.image_canonical, action.architecture
        )
        return ClearImageCustomResourceLimitActionResult(image_data=image_data)



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.agent.actions.get_total_resources import (
    GetTotalResourcesAction,
    GetTotalResourcesActionResult,
)
from ai.backend.manager.services.agent.actions.get_watcher_status import (
    GetWatcherStatusAction,
    GetWatcherStatusActionResult,
)
from ai.backend.manager.services.agent.actions.handle_heartbeat import (
    HandleHeartbeatAction,
    HandleHeartbeatActionResult,
)
from ai.backend.manager.services.agent.actions.mark_agent_exit import (
    MarkAgentExitAction,
    MarkAgentExitActionResult,
)
from ai.backend.manager.services.agent.actions.mark_agent_running import (
    MarkAgentRunningAction,
    MarkAgentRunningActionResult,
)
from ai.backend.manager.services.agent.actions.recalculate_usage import (
    RecalculateUsageAction,
    RecalculateUsageActionResult,
)
from ai.backend.manager.services.agent.actions.remove_agent_from_images import (
    RemoveAgentFromImagesAction,
    RemoveAgentFromImagesActionResult,
)
from ai.backend.manager.services.agent.actions.sync_agent_registry import (
    SyncAgentRegistryAction,
    SyncAgentRegistryActionResult,
)
from ai.backend.manager.services.agent.actions.watcher_agent_restart import (
    WatcherAgentRestartAction,
    WatcherAgentRestartActionResult,
)
from ai.backend.manager.services.agent.actions.watcher_agent_start import (
    WatcherAgentStartAction,
    WatcherAgentStartActionResult,
)
from ai.backend.manager.services.agent.actions.watcher_agent_stop import (
    WatcherAgentStopAction,
    WatcherAgentStopActionResult,
)
from ai.backend.manager.services.agent.service import AgentService


class AgentProcessors(AbstractProcessorPackage):
    sync_agent_registry: ActionProcessor[SyncAgentRegistryAction, SyncAgentRegistryActionResult]
    get_watcher_status: ActionProcessor[GetWatcherStatusAction, GetWatcherStatusActionResult]
    watcher_agent_start: ActionProcessor[WatcherAgentStartAction, WatcherAgentStartActionResult]
    watcher_agent_restart: ActionProcessor[
        WatcherAgentRestartAction, WatcherAgentRestartActionResult
    ]
    watcher_agent_stop: ActionProcessor[WatcherAgentStopAction, WatcherAgentStopActionResult]
    recalculate_usage: ActionProcessor[RecalculateUsageAction, RecalculateUsageActionResult]
    get_total_resources: ActionProcessor[GetTotalResourcesAction, GetTotalResourcesActionResult]
    handle_heartbeat: ActionProcessor[HandleHeartbeatAction, HandleHeartbeatActionResult]
    mark_agent_exit: ActionProcessor[MarkAgentExitAction, MarkAgentExitActionResult]
    mark_agent_running: ActionProcessor[MarkAgentRunningAction, MarkAgentRunningActionResult]
    remove_agent_from_images: ActionProcessor[
        RemoveAgentFromImagesAction, RemoveAgentFromImagesActionResult
    ]

    def __init__(self, service: AgentService, action_monitors: list[ActionMonitor]) -> None:
        self.sync_agent_registry = ActionProcessor(service.sync_agent_registry, action_monitors)
        self.get_watcher_status = ActionProcessor(service.get_watcher_status, action_monitors)
        self.watcher_agent_start = ActionProcessor(service.watcher_agent_start, action_monitors)
        self.watcher_agent_restart = ActionProcessor(service.watcher_agent_restart, action_monitors)
        self.watcher_agent_stop = ActionProcessor(service.watcher_agent_stop, action_monitors)
        self.recalculate_usage = ActionProcessor(service.recalculate_usage, action_monitors)
        self.get_total_resources = ActionProcessor(service.get_total_resources, action_monitors)
        self.handle_heartbeat = ActionProcessor(service.handle_heartbeat, action_monitors)
        self.mark_agent_exit = ActionProcessor(service.mark_agent_exit, action_monitors)
        self.mark_agent_running = ActionProcessor(service.mark_agent_running, action_monitors)
        self.remove_agent_from_images = ActionProcessor(
            service.remove_agent_from_images, action_monitors
        )

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            SyncAgentRegistryAction.spec(),
            GetWatcherStatusAction.spec(),
            WatcherAgentStartAction.spec(),
            WatcherAgentRestartAction.spec(),
            WatcherAgentStopAction.spec(),
            RecalculateUsageAction.spec(),
            GetTotalResourcesAction.spec(),
            HandleHeartbeatAction.spec(),
        ]



from __future__ import annotations

import enum
import functools
import logging
import uuid
from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from decimal import Decimal
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    Optional,
    Self,
    Tuple,
    TypeAlias,
    cast,
    override,
)
from uuid import UUID

import sqlalchemy as sa
import trafaret as t
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncSession
from sqlalchemy.orm import foreign, joinedload, load_only, relationship, selectinload
from sqlalchemy.sql.expression import true

from ai.backend.common.container_registry import ContainerRegistryType
from ai.backend.common.docker import ImageRef
from ai.backend.common.exception import UnknownImageReference
from ai.backend.common.types import (
    AutoPullBehavior,
    BinarySize,
    ImageAlias,
    ImageConfig,
    ImageRegistry,
    ResourceSlot,
    SlotName,
)
from ai.backend.common.utils import join_non_empty
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.loader.legacy_etcd_loader import LegacyEtcdLoader
from ai.backend.manager.data.image.types import (
    ImageAliasData,
    ImageData,
    ImageIdentifier,
    ImageLabelsData,
    ImageResourcesData,
    ImageStatus,
    ImageType,
    RescanImagesResult,
)
from ai.backend.manager.defs import INTRINSIC_SLOTS, INTRINSIC_SLOTS_MIN

from ..container_registry import get_container_registry_cls
from ..errors.image import ImageNotFound
from ..models.container_registry import ContainerRegistryRow
from .base import (
    GUID,
    Base,
    ForeignKeyIDColumn,
    IDColumn,
    StrEnumType,
    StructuredJSONColumn,
)
from .rbac import (
    AbstractPermissionContext,
    AbstractPermissionContextBuilder,
    DomainScope,
    ProjectScope,
    ScopeType,
    UserScope,
    get_predefined_roles_in_scope,
)
from .rbac.context import ClientContext
from .rbac.exceptions import InvalidScope
from .rbac.permission_defs import ImagePermission
from .user import UserRole, UserRow
from .utils import ExtendedAsyncSAEngine

if TYPE_CHECKING:
    from ai.backend.common.bgtask.bgtask import ProgressReporter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


__all__ = (
    "rescan_images",
    "ImageAliasRow",
    "ImageLoadFilter",
    "ImageRow",
    "ImageIdentifier",
    "PublicImageLoadFilter",
)


class PublicImageLoadFilter(enum.StrEnum):
    """Shorthand of `ImageLoadFilter` enum with `CUSTOMIZED_GLOBAL` removed (as it is not intended for API input)."""

    GENERAL = "general"
    """Include general purpose images."""
    OPERATIONAL = "operational"
    """Include operational images."""
    CUSTOMIZED = "customized"
    """Include customized images owned or accessible by API callee."""


class ImageLoadFilter(enum.StrEnum):
    """Enum describing kind of a "search preset" when loading Image data via GQL. Not intended for declaring attributes of image data itself."""

    GENERAL = "general"
    """Include general purpose images."""
    OPERATIONAL = "operational"
    """Include operational images."""
    CUSTOMIZED = "customized"
    """Include customized images owned or accessible by API callee."""
    CUSTOMIZED_GLOBAL = "customized-global"
    """Include every customized images filed at the system. Effective only for superadmin. CUSTOMIZED and CUSTOMIZED_GLOBAL are mutually exclusive."""


class RelationLoadingOption(enum.StrEnum):
    ALIASES = enum.auto()
    ENDPOINTS = enum.auto()
    REGISTRY = enum.auto()


def _apply_loading_option(
    query_stmt: sa.sql.Select, options: Iterable[RelationLoadingOption]
) -> sa.sql.Select:
    for op in options:
        match op:
            case RelationLoadingOption.ALIASES:
                query_stmt = query_stmt.options(selectinload(ImageRow.aliases))
            case RelationLoadingOption.REGISTRY:
                query_stmt = query_stmt.options(joinedload(ImageRow.registry_row))
            case RelationLoadingOption.ENDPOINTS:
                query_stmt = query_stmt.options(selectinload(ImageRow.endpoints))
    return query_stmt


async def load_configured_registries(
    db: ExtendedAsyncSAEngine,
    project: Optional[str],
) -> dict[str, ContainerRegistryRow]:
    join = functools.partial(join_non_empty, sep="/")

    async with db.begin_readonly_session() as session:
        result = await session.execute(sa.select(ContainerRegistryRow))
        if project:
            registries = cast(
                dict[str, ContainerRegistryRow],
                {
                    join(row.registry_name, row.project): row
                    for row in result.scalars().all()
                    if row.project == project
                },
            )
        else:
            registries = cast(
                dict[str, ContainerRegistryRow],
                {join(row.registry_name, row.project): row for row in result.scalars().all()},
            )

    return cast(dict[str, ContainerRegistryRow], registries)


async def scan_registries(
    db: ExtendedAsyncSAEngine,
    registries: dict[str, ContainerRegistryRow],
    reporter: Optional[ProgressReporter] = None,
) -> RescanImagesResult:
    """
    Performs an image rescan for all images in the registries.
    """
    from ai.backend.manager.data.image.types import RescanImagesResult

    images, errors = [], []

    for registry_key, registry_row in registries.items():
        registry_name = ImageRef.parse_image_str(registry_key, "*").registry
        log.info('Scanning kernel images from the registry "{0}"', registry_name)

        scanner_cls = get_container_registry_cls(registry_row)
        scanner = scanner_cls(db, registry_name, registry_row)

        try:
            scan_result = await scanner.rescan_single_registry(reporter)
            images.extend(scan_result.images or [])
            errors.extend(scan_result.errors or [])
        except Exception as e:
            errors.append(str(e))

    return RescanImagesResult(images=images, errors=errors)


async def scan_single_image(
    db: ExtendedAsyncSAEngine,
    registry_key: str,
    registry_row: ContainerRegistryRow,
    image_canonical: str,
) -> RescanImagesResult:
    """
    Performs a scan for a single image.
    """
    registry_name = ImageRef.parse_image_str(registry_key, "*").registry
    image_name = image_canonical.removeprefix(registry_name + "/")

    log.debug("running a per-image metadata scan: {}, {}", registry_name, image_name)

    scanner_cls = get_container_registry_cls(registry_row)
    scanner = scanner_cls(db, registry_name, registry_row)
    return await scanner.scan_single_ref(image_name)


def filter_registry_dict(
    registries: dict[str, ContainerRegistryRow],
    condition: Callable[[str, ContainerRegistryRow], bool],
) -> dict[str, ContainerRegistryRow]:
    return {
        registry_key: registry_row
        for registry_key, registry_row in registries.items()
        if condition(registry_key, registry_row)
    }


def filter_registries_by_img_canonical(
    registries: dict[str, ContainerRegistryRow], registry_or_image: str
) -> dict[str, ContainerRegistryRow]:
    """
    Filters the matching registry assuming `registry_or_image` is an image canonical name.
    """
    return filter_registry_dict(
        registries,
        lambda registry_key, _row: registry_or_image.startswith(registry_key + "/"),
    )


def filter_registries_by_registry_name(
    registries: dict[str, ContainerRegistryRow], registry_or_image: str
) -> dict[str, ContainerRegistryRow]:
    """
    Filters the matching registry assuming `registry_or_image` is a registry name.
    """
    return filter_registry_dict(
        registries,
        lambda registry_key, _row: registry_key.startswith(registry_or_image),
    )


async def rescan_images(
    db: ExtendedAsyncSAEngine,
    registry_or_image: Optional[str] = None,
    project: Optional[str] = None,
    *,
    reporter: Optional[ProgressReporter] = None,
) -> RescanImagesResult:
    """
    Rescan container registries and the update images table.
    Refer to the comments below for details on the function's behavior.

    If registry name is provided for `registry_or_image`, scans all images in the specified registry.
    If image canonical name is provided for `registry_or_image`, only scan the image.
    If the `registry_or_image` is not provided, scan all configured registries.

    If `project` is provided, only scan the registries associated with the project.
    """
    registries = await load_configured_registries(db, project)

    if registry_or_image is None:
        return await scan_registries(db, registries, reporter=reporter)

    matching_registries = filter_registries_by_img_canonical(registries, registry_or_image)

    if matching_registries:
        if len(matching_registries) > 1:
            raise RuntimeError(
                "ContainerRegistryRows exist with the same registry_name and project!",
            )

        registry_key, registry_row = next(iter(matching_registries.items()))
        return await scan_single_image(db, registry_key, registry_row, registry_or_image)

    matching_registries = filter_registries_by_registry_name(registries, registry_or_image)

    if not matching_registries:
        raise RuntimeError("It is an unknown registry.", registry_or_image)

    log.debug("running a per-registry metadata scan")
    return await scan_registries(db, matching_registries, reporter=reporter)
    # TODO: delete images removed from registry?


type Resources = dict[SlotName, dict[str, Any]]


# Defined for avoiding circular import
def _get_image_endpoint_join_condition():
    from ai.backend.manager.models.endpoint import EndpointRow

    return ImageRow.id == foreign(EndpointRow.image)


class ImageRow(Base):
    __tablename__ = "images"
    __table_args__ = (
        sa.UniqueConstraint(
            "registry", "project", "name", "tag", "architecture", name="uq_image_identifier"
        ),
    )

    id = IDColumn("id")
    name = sa.Column("name", sa.String, nullable=False, index=True)
    project = sa.Column("project", sa.String, nullable=True)
    image = sa.Column("image", sa.String, nullable=False, index=True)
    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        index=True,
    )
    tag = sa.Column("tag", sa.TEXT)
    registry = sa.Column("registry", sa.String, nullable=False, index=True)
    registry_id = sa.Column("registry_id", GUID, nullable=False, index=True)
    architecture = sa.Column(
        "architecture", sa.String, nullable=False, index=True, default="x86_64"
    )
    config_digest = sa.Column("config_digest", sa.CHAR(length=72), nullable=False)
    size_bytes = sa.Column("size_bytes", sa.BigInteger, nullable=False)
    is_local = sa.Column(
        "is_local",
        sa.Boolean,
        nullable=False,
        server_default=sa.sql.expression.false(),
    )
    type = sa.Column("type", sa.Enum(ImageType), nullable=False)
    accelerators = sa.Column("accelerators", sa.String)
    labels = sa.Column("labels", sa.JSON, nullable=False, default=dict)
    _resources = sa.Column(
        "resources",
        StructuredJSONColumn(
            t.Mapping(
                t.String,
                t.Dict({
                    t.Key("min"): t.String,
                    t.Key("max", default=None): t.Null | t.String,
                }),
            ),
        ),
        nullable=False,
    )  # Custom resource limits designated by the user

    status = sa.Column(
        "status",
        StrEnumType(ImageStatus),
        default=ImageStatus.ALIVE,
        server_default=ImageStatus.ALIVE.name,
        nullable=False,
    )

    aliases: relationship
    # sessions = relationship("SessionRow", back_populates="image_row")
    endpoints = relationship(
        "EndpointRow",
        primaryjoin=_get_image_endpoint_join_condition,
        back_populates="image_row",
    )

    registry_row = relationship(
        "ContainerRegistryRow",
        back_populates="image_rows",
        primaryjoin="ContainerRegistryRow.id == foreign(ImageRow.registry_id)",
    )

    def __init__(
        self,
        name,
        project,
        architecture,
        registry_id,
        is_local=False,
        registry=None,
        image=None,
        tag=None,
        config_digest=None,
        size_bytes=None,
        type=None,
        accelerators=None,
        labels=None,
        resources=None,
        status=ImageStatus.ALIVE,
    ) -> None:
        self.name = name
        self.project = project
        self.registry = registry
        self.registry_id = registry_id
        self.image = image
        self.tag = tag
        self.architecture = architecture
        self.is_local = is_local
        self.config_digest = config_digest
        self.size_bytes = size_bytes
        self.type = type
        self.accelerators = accelerators
        self.labels = labels
        self._resources = resources
        self.status = status

    @property
    def trimmed_digest(self) -> str:
        return self.config_digest.strip()

    @property
    def image_ref(self) -> ImageRef:
        # Empty image name
        if self.project == self.image:
            image_name = ""
            _, tag = ImageRef.parse_image_tag(self.name.split(f"{self.registry}/", maxsplit=1)[1])
        else:
            join = functools.partial(join_non_empty, sep="/")
            image_and_tag = self.name.removeprefix(f"{join(self.registry, self.project)}/")
            image_name, tag = ImageRef.parse_image_tag(image_and_tag)

        return ImageRef(
            image_name, self.project, tag, self.registry, self.architecture, self.is_local
        )

    @property
    def customized(self) -> bool:
        return self.labels.get("ai.backend.customized-image.owner") is not None

    @classmethod
    async def from_alias(
        cls,
        session: AsyncSession,
        alias: str,
        load_aliases: bool = False,
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
        *,
        loading_options: Iterable[RelationLoadingOption] = tuple(),
    ) -> Self:
        query = (
            sa.select(ImageRow)
            .select_from(ImageRow)
            .join(ImageAliasRow, ImageRow.aliases.and_(ImageAliasRow.alias == alias))
        )
        if load_aliases:
            query = query.options(selectinload(ImageRow.aliases))
        if filter_by_statuses:
            query = query.where(ImageRow.status.in_(filter_by_statuses))

        query = _apply_loading_option(query, loading_options)
        result = await session.scalar(query)
        if result is not None:
            return result
        else:
            raise UnknownImageReference

    @classmethod
    async def from_image_identifier(
        cls,
        session: AsyncSession,
        identifier: ImageIdentifier,
        load_aliases: bool = True,
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
        *,
        loading_options: Iterable[RelationLoadingOption] = tuple(),
    ) -> Self:
        query = sa.select(ImageRow).where(
            (ImageRow.name == identifier.canonical)
            & (ImageRow.architecture == identifier.architecture)
        )

        if load_aliases:
            query = query.options(selectinload(ImageRow.aliases))
        if filter_by_statuses:
            query = query.where(ImageRow.status.in_(filter_by_statuses))

        query = _apply_loading_option(query, loading_options)

        result = await session.execute(query)
        candidates: List[ImageRow] = result.scalars().all()

        if len(candidates) <= 0:
            raise UnknownImageReference(identifier.canonical)

        return candidates[0]

    @classmethod
    async def from_image_ref(
        cls,
        session: AsyncSession,
        ref: ImageRef,
        *,
        strict_arch: bool = False,
        load_aliases: bool = False,
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
        loading_options: Iterable[RelationLoadingOption] = tuple(),
    ) -> Self:
        """
        Loads a image row that corresponds to the given ImageRef object.

        When *strict_arch* is False and the image table has only one row
        with respect to requested canonical, this function will
        return that row regardless of the image architecture.
        """
        query = sa.select(ImageRow).where(ImageRow.name == ref.canonical)
        if load_aliases:
            query = query.options(selectinload(ImageRow.aliases))
        if filter_by_statuses:
            query = query.where(ImageRow.status.in_(filter_by_statuses))

        query = _apply_loading_option(query, loading_options)

        result = await session.execute(query)
        candidates: List[ImageRow] = result.scalars().all()

        if len(candidates) == 0:
            raise UnknownImageReference(ref)
        if len(candidates) == 1 and not strict_arch:
            return candidates[0]
        for row in candidates:
            if row.architecture == ref.architecture:
                return row
        raise UnknownImageReference(ref)

    @classmethod
    def from_dataclass(cls, image_data: ImageData) -> Self:
        image_row = ImageRow(
            name=image_data.name,
            project=image_data.project,
            image=image_data.image,
            tag=image_data.tag,
            registry=image_data.registry,
            registry_id=image_data.registry_id,
            architecture=image_data.architecture,
            config_digest=image_data.config_digest,
            size_bytes=image_data.size_bytes,
            is_local=image_data.is_local,
            type=image_data.type,
            accelerators=image_data.accelerators,
            labels=image_data.labels.label_data,
            resources=image_data.resources.resources_data,
            status=image_data.status,
        )
        image_row.id = image_data.id
        image_row.created_at = image_data.created_at
        return image_row

    @classmethod
    def from_optional_dataclass(cls, image_data: Optional[ImageData]) -> Optional[Self]:
        if image_data is None:
            return None
        return cls.from_dataclass(image_data)

    @classmethod
    async def resolve(
        cls,
        session: AsyncSession,
        reference_candidates: list[ImageAlias | ImageRef | ImageIdentifier],
        *,
        strict_arch: bool = False,
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
        load_aliases: bool = True,
        loading_options: Iterable[RelationLoadingOption] = tuple(),
    ) -> Self:
        """
        Resolves a matching row in the image table from image references and/or aliases.
        If candidate element is `ImageRef`, this method will try to resolve image with matching
        `ImageRef` description. Otherwise, if element is `str`, this will try to follow the alias.
        If multiple elements are supplied, this method will return the first matched `ImageRow`
        among those elements.
        Passing the canonical image reference as string directly to resolve image data
        is no longer possible. You need to declare ImageRef object explicitly if you're using string
        as an canonical image references. For example:
        .. code-block::
           await ImageRow.resolve(
               conn,
               [
                   ImageRef(
                       image_name,
                       project,
                       registry,
                       tag,
                       architecture,
                       is_local,
                   ),
                   ImageIdentifier(canonical, architecture),
                   ImageAlias(image_alias),
               ],
           )

        When *strict_arch* is False and the image table has only one row
        with respect to requested canonical, this function will
        return that row regardless of the image architecture.

        When *load_aliases* is True, it tries to resolve the alias chain.
        Otherwise it finds only the direct image references.
        """
        searched_refs = []
        for reference in reference_candidates:
            resolver_func: Any = None
            if isinstance(reference, str):
                resolver_func = cls.from_alias
                searched_refs.append(f"alias:{reference!r}")
            elif isinstance(reference, ImageRef):
                resolver_func = functools.partial(cls.from_image_ref, strict_arch=strict_arch)
                searched_refs.append(f"ref:{reference.canonical!r}")
            elif isinstance(reference, ImageIdentifier):
                resolver_func = cls.from_image_identifier
                searched_refs.append(f"identifier:{reference!r}")
            try:
                if row := await resolver_func(
                    session,
                    reference,
                    load_aliases=load_aliases,
                    filter_by_statuses=filter_by_statuses,
                    loading_options=loading_options,
                ):
                    return row
            except UnknownImageReference:
                continue
        raise ImageNotFound("Unknown image references: " + ", ".join(searched_refs))

    @classmethod
    async def lookup(
        cls,
        session: AsyncSession,
        image_identifier: ImageIdentifier,
    ) -> Self:
        """
        Lookup ImageRow by ImageIdentifier, also trying canonical as alias.

        Args:
            session: Database session
            image_identifier: ImageIdentifier containing canonical name and architecture

        Returns:
            ImageRow instance

        Raises:
            ImageNotFound: If no matching image is found
        """
        identifiers: list[ImageAlias | ImageRef | ImageIdentifier] = [
            image_identifier,
            ImageAlias(image_identifier.canonical),
        ]

        return await cls.resolve(session, identifiers)

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        image_id: UUID,
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
        load_aliases: bool = False,
    ) -> Optional[Self]:
        query = sa.select(ImageRow).where(ImageRow.id == image_id)
        if load_aliases:
            query = query.options(selectinload(ImageRow.aliases))
        if filter_by_statuses:
            query = query.where(ImageRow.status.in_(filter_by_statuses))

        result = await session.execute(query)
        return result.scalar()

    @classmethod
    async def list(
        cls,
        session: AsyncSession,
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
        load_aliases: bool = False,
    ) -> list[Self]:
        query = sa.select(ImageRow)
        if load_aliases:
            query = query.options(selectinload(ImageRow.aliases))
        if filter_by_statuses:
            query = query.where(ImageRow.status.in_(filter_by_statuses))

        result = await session.execute(query)
        return result.scalars().all()

    def __str__(self) -> str:
        return self.image_ref.canonical + f" ({self.image_ref.architecture})"

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def resources(self) -> Resources:
        custom_resources = self._resources or {}
        label_resources = self.get_resources_from_labels()

        merged_resources: dict[str, Any] = {}

        for label_key in {*custom_resources.keys(), *label_resources.keys()}:
            custom_spec = custom_resources.get(label_key, {})
            label_spec = label_resources.get(label_key, {})

            merged_spec = dict(custom_spec)
            for label_spec_key, value in label_spec.items():
                merged_spec.setdefault(label_spec_key, value)

            # TODO: Consider other slot types
            if label_key in INTRINSIC_SLOTS.keys():
                mins = [m for m in (custom_spec.get("min"), label_spec.get("min")) if m is not None]
                if mins:
                    match label_key:
                        case "cpu":
                            merged_spec["min"] = max(mins)
                        case "mem":
                            merged_spec["min"] = max(mins, key=BinarySize.from_str)

            merged_resources[label_key] = merged_spec
        return ImageRow._resources.type._schema.check(merged_resources)

    def get_resources_from_labels(self) -> Resources:
        if self.labels is None:
            raise RuntimeError(f'Labels not loaded in the ImageRow "({self.image_ref.canonical})"')

        RESOURCE_LABEL_PREFIX = "ai.backend.resource.min."

        resources = {  # default fallback if not defined
            "cpu": {"min": INTRINSIC_SLOTS_MIN[SlotName("cpu")], "max": None},
            "mem": {"min": INTRINSIC_SLOTS_MIN[SlotName("mem")], "max": None},
        }

        for k, v in filter(
            lambda pair: pair[0].startswith(RESOURCE_LABEL_PREFIX), self.labels.items()
        ):
            res_key = k[len(RESOURCE_LABEL_PREFIX) :]
            resources[res_key] = {"min": v}

        return ImageRow._resources.type._schema.check(resources)

    async def get_min_slot(self, etcd_loader: LegacyEtcdLoader) -> ResourceSlot:
        slot_units = await etcd_loader.get_resource_slots()
        min_slot = ResourceSlot()

        for slot_key, resource in self.resources.items():
            slot_unit = slot_units.get(slot_key)
            if slot_unit is None:
                # ignore unknown slots
                continue
            min_value = resource.get("min")
            if min_value is None:
                min_value = Decimal(0)
            if slot_unit == "bytes":
                if not isinstance(min_value, Decimal):
                    min_value = BinarySize.from_str(min_value)
            else:
                if not isinstance(min_value, Decimal):
                    min_value = Decimal(min_value)
            min_slot[slot_key] = min_value

        # fill missing
        for slot_key in slot_units.keys():
            if slot_key not in min_slot:
                min_slot[slot_key] = Decimal(0)

        return min_slot

    def _parse_row(self):
        res_limits = []
        for slot_key, slot_range in self.resources.items():
            min_value = slot_range.get("min")
            if min_value is None:
                min_value = Decimal(0)
            max_value = slot_range.get("max")
            if max_value is None:
                max_value = Decimal("Infinity")
            res_limits.append({
                "key": slot_key,
                "min": min_value,
                "max": max_value,
            })

        accels = self.accelerators
        if accels is None:
            accels = []
        else:
            accels = accels.split(",")

        return {
            "canonical_ref": self.name,
            "name": self.image,
            "humanized_name": self.image,  # TODO: implement
            "tag": self.tag,
            "architecture": self.architecture,
            "registry": self.registry,
            "digest": self.trimmed_digest or None,
            "labels": self.labels,
            "size_bytes": self.size_bytes,
            "resource_limits": res_limits,
            "supported_accelerators": accels,
        }

    async def inspect(self) -> Mapping[str, Any]:
        parsed_image_info = self._parse_row()
        parsed_image_info["reverse_aliases"] = [x.alias for x in self.aliases]
        return parsed_image_info

    async def mark_as_deleted(self, db_session: AsyncSession) -> None:
        self.status = ImageStatus.DELETED
        await db_session.flush()

    def set_resource_limit(
        self,
        slot_type: str,
        value_range: Tuple[Optional[Decimal], Optional[Decimal]],
    ):
        resources = self._resources
        if resources.get(slot_type) is None:
            resources[slot_type] = {}
        if value_range[0] is not None:
            resources[slot_type]["min"] = str(value_range[0])
        if value_range[1] is not None:
            resources[slot_type]["max"] = str(value_range[1])

        self._resources = resources

    def is_owned_by(self, user_id: UUID) -> bool:
        if not self.customized:
            return False
        return self.labels["ai.backend.customized-image.owner"].split(":")[1] == str(user_id)

    def to_dataclass(self) -> ImageData:
        from ai.backend.manager.data.image.types import ImageData

        return ImageData(
            id=self.id,
            name=self.name,
            project=self.project,
            image=self.image,
            created_at=self.created_at,
            tag=self.tag,
            registry=self.registry,
            registry_id=self.registry_id,
            architecture=self.architecture,
            config_digest=self.config_digest,
            size_bytes=self.size_bytes,
            is_local=self.is_local,
            type=self.type,
            accelerators=self.accelerators,
            labels=ImageLabelsData(label_data=self.labels),
            resources=ImageResourcesData(resources_data=self.resources),
            status=self.status,
        )

    async def untag_image_from_registry(
        self, db: ExtendedAsyncSAEngine, session: AsyncSession
    ) -> None:
        """
        Works only for HarborV2 registries.
        """
        from ai.backend.manager.container_registry.harbor import HarborRegistry_v2

        query = sa.select(ContainerRegistryRow).where(ContainerRegistryRow.id == self.registry_id)

        registry_info: ContainerRegistryRow = (await session.execute(query)).scalar()
        if registry_info.type != ContainerRegistryType.HARBOR2:
            raise NotImplementedError("This feature is only supported for Harbor 2 registries")

        scanner = HarborRegistry_v2(db, self.image_ref.registry, registry_info)
        await scanner.untag(self.image_ref)


async def bulk_get_image_configs(
    image_refs: Iterable[ImageRef],
    auto_pull: AutoPullBehavior = AutoPullBehavior.DIGEST,
    *,
    db_session: AsyncSession,
) -> list[ImageConfig]:
    result: list[ImageConfig] = []

    for ref in image_refs:
        resolved_image_info = await ImageRow.resolve(db_session, [ref])

        registry_info: ImageRegistry
        if resolved_image_info.image_ref.is_local:
            registry_info = {
                "name": ref.registry,
                "url": "http://127.0.0.1",  # "http://localhost",
                "username": None,
                "password": None,
            }
        else:
            url, credential = await ContainerRegistryRow.get_container_registry_info(
                db_session, resolved_image_info.registry_id
            )
            registry_info = {
                "name": ref.registry,
                "url": str(url),
                "username": credential["username"],
                "password": credential["password"],
            }

        image_conf: ImageConfig = {
            "architecture": ref.architecture,
            "project": resolved_image_info.project,
            "canonical": ref.canonical,
            "is_local": resolved_image_info.image_ref.is_local,
            "digest": resolved_image_info.trimmed_digest,
            "labels": resolved_image_info.labels,
            "repo_digest": None,
            "registry": registry_info,
            "auto_pull": auto_pull,
        }

        result.append(image_conf)

    return result


class ImageAliasRow(Base):
    __tablename__ = "image_aliases"
    id = IDColumn("id")
    alias = sa.Column("alias", sa.String, unique=True, index=True)
    image_id = ForeignKeyIDColumn("image", "images.id", nullable=False)
    image: relationship

    @classmethod
    async def create(
        cls,
        session: AsyncSession,
        alias: str,
        target: ImageRow,
    ) -> Self:
        existing_alias: Optional[ImageRow] = await session.scalar(
            sa.select(ImageAliasRow)
            .where(ImageAliasRow.alias == alias)
            .options(selectinload(ImageAliasRow.image)),
        )
        if existing_alias is not None:
            raise ValueError(
                f"alias already created with ({existing_alias.image})",
            )
        new_alias = ImageAliasRow(
            alias=alias,
            image_id=target.id,
        )
        session.add_all([new_alias])
        return new_alias

    @classmethod
    def from_dataclass(cls, alias_data: ImageAliasData, image_id: uuid.UUID) -> Self:
        return cls(id=alias_data.id, alias=alias_data.alias, image_id=image_id)

    def to_dataclass(self) -> ImageAliasData:
        return ImageAliasData(id=self.id, alias=self.alias)


ImageRow.aliases = relationship("ImageAliasRow", back_populates="image")
ImageAliasRow.image = relationship("ImageRow", back_populates="aliases")


WhereClauseType: TypeAlias = (
    sa.sql.expression.BinaryExpression | sa.sql.expression.BooleanClauseList
)
# TypeAlias is deprecated since 3.12 but mypy does not follow up yet

ALL_IMAGE_PERMISSIONS = frozenset([perm for perm in ImagePermission])
OWNER_PERMISSIONS: frozenset[ImagePermission] = ALL_IMAGE_PERMISSIONS
ADMIN_PERMISSIONS: frozenset[ImagePermission] = frozenset([
    ImagePermission.READ_ATTRIBUTE,
    ImagePermission.CREATE_CONTAINER,
])
MONITOR_PERMISSIONS: frozenset[ImagePermission] = frozenset([
    ImagePermission.READ_ATTRIBUTE,
    ImagePermission.UPDATE_ATTRIBUTE,
])
PRIVILEGED_MEMBER_PERMISSIONS: frozenset[ImagePermission] = frozenset([
    ImagePermission.READ_ATTRIBUTE,
    ImagePermission.CREATE_CONTAINER,
])
MEMBER_PERMISSIONS: frozenset[ImagePermission] = frozenset([
    ImagePermission.READ_ATTRIBUTE,
    ImagePermission.CREATE_CONTAINER,
])


@dataclass
class ImagePermissionContext(AbstractPermissionContext[ImagePermission, ImageRow, UUID]):
    @property
    def query_condition(self) -> Optional[WhereClauseType]:
        cond: WhereClauseType | None = None

        def _OR_coalesce(
            base_cond: Optional[WhereClauseType],
            _cond: sa.sql.expression.BinaryExpression,
        ) -> WhereClauseType:
            return base_cond | _cond if base_cond is not None else _cond

        if self.object_id_to_additional_permission_map:
            cond = _OR_coalesce(
                cond, ImageRow.id.in_(self.object_id_to_additional_permission_map.keys())
            )
        if self.object_id_to_overriding_permission_map:
            cond = _OR_coalesce(
                cond, ImageRow.id.in_(self.object_id_to_overriding_permission_map.keys())
            )
        return cond

    async def build_query(self) -> Optional[sa.sql.Select]:
        cond = self.query_condition
        if cond is None:
            return None
        return sa.select(ImageRow).where(cond)

    async def calculate_final_permission(self, rbac_obj: ImageRow) -> frozenset[ImagePermission]:
        image_row = rbac_obj
        image_id = cast(UUID, image_row.id)
        permissions: set[ImagePermission] = set()

        if (
            overriding_perm := self.object_id_to_overriding_permission_map.get(image_id)
        ) is not None:
            permissions = set(overriding_perm)
        else:
            permissions |= self.object_id_to_additional_permission_map.get(image_id, set())

        return frozenset(permissions)


@dataclass
class ImageAccessCriteria:
    user_id: UUID
    allowed_registries: set[str]

    def is_accessible_image(self, image: ImageRow) -> bool:
        """
        Check if the image is accessible by the user.
        """
        if image.registry not in self.allowed_registries:
            return False
        if image.customized and not image.is_owned_by(self.user_id):
            return False
        return True


class ImagePermissionContextBuilder(
    AbstractPermissionContextBuilder[ImagePermission, ImagePermissionContext]
):
    db_session: AsyncSession

    def __init__(self, db_session: AsyncSession) -> None:
        self.db_session = db_session

    @override
    async def calculate_permission(
        self,
        ctx: ClientContext,
        target_scope: ScopeType,
    ) -> frozenset[ImagePermission]:
        roles = await get_predefined_roles_in_scope(ctx, target_scope, self.db_session)
        permissions = await self._calculate_permission_by_predefined_roles(roles)
        permissions |= await self.apply_customized_role(ctx, target_scope)
        return permissions

    async def apply_customized_role(
        self,
        ctx: ClientContext,
        target_scope: ScopeType,
    ) -> frozenset[ImagePermission]:
        if ctx.user_role == UserRole.SUPERADMIN:
            return ALL_IMAGE_PERMISSIONS
        return frozenset()

    @override
    async def build_ctx_in_system_scope(
        self,
        ctx: ClientContext,
    ) -> ImagePermissionContext:
        perm_ctx = ImagePermissionContext()
        user_accessible_project_scopes = await self._get_user_accessible_project_scopes(
            ctx, UserScope(ctx.user_id)
        )
        global_project_scopes_perm_ctx = await self._in_project_scopes_global(
            ctx, user_accessible_project_scopes
        )
        perm_ctx.merge(global_project_scopes_perm_ctx)
        non_global_project_scopes_perm_ctx = await self._in_project_scopes_non_global(
            ctx, user_accessible_project_scopes
        )
        perm_ctx.merge(non_global_project_scopes_perm_ctx)

        user_scope_perm_ctx = await self._in_user_scope(
            ctx,
            UserScope(ctx.user_id),
        )

        perm_ctx.merge(user_scope_perm_ctx)
        return perm_ctx

    @override
    async def build_ctx_in_domain_scope(
        self,
        ctx: ClientContext,
        scope: DomainScope,
    ) -> ImagePermissionContext:
        perm_ctx = ImagePermissionContext()
        domain_scope_perm_ctx = await self._in_domain_scope(ctx, scope)
        perm_ctx.merge(domain_scope_perm_ctx)
        user_scope_perm_ctx = await self._in_user_scope(ctx, UserScope(ctx.user_id))
        perm_ctx.merge(user_scope_perm_ctx)

        project_scopes = await self._get_domain_accessible_project_scopes(ctx, scope)
        non_global_container_registries_perm_ctx = await self._in_project_scopes_non_global(
            ctx, project_scopes
        )
        perm_ctx.merge(non_global_container_registries_perm_ctx)
        return perm_ctx

    @override
    async def build_ctx_in_project_scope(
        self,
        ctx: ClientContext,
        scope: ProjectScope,
    ) -> ImagePermissionContext:
        perm_ctx = ImagePermissionContext()
        global_container_registries_perm_ctx = await self._in_project_scopes_global(ctx, [scope])
        perm_ctx.merge(global_container_registries_perm_ctx)
        non_global_container_registries_perm_ctx = await self._in_project_scopes_non_global(
            ctx, [scope]
        )
        perm_ctx.merge(non_global_container_registries_perm_ctx)
        return perm_ctx

    @override
    async def build_ctx_in_user_scope(
        self,
        ctx: ClientContext,
        scope: UserScope,
    ) -> ImagePermissionContext:
        perm_ctx = ImagePermissionContext()
        user_scope_perm_ctx = await self._in_user_scope(ctx, scope)
        perm_ctx.merge(user_scope_perm_ctx)

        project_scopes = await self._get_user_accessible_project_scopes(ctx, scope)

        # We should fetch only customized images
        non_global_container_registries_perm_ctx = await self._in_project_scopes_non_global(
            ctx, project_scopes
        )
        perm_ctx.merge(non_global_container_registries_perm_ctx)
        return perm_ctx

    async def _get_allowed_registries_for_user(self, ctx: ClientContext, user_id: UUID) -> set[str]:
        user_query_stmt = (
            sa.select(UserRow).where(UserRow.uuid == user_id).options(joinedload(UserRow.domain))
        )
        user_row = cast(Optional[UserRow], await self.db_session.scalar(user_query_stmt))
        if user_row is None:
            raise InvalidScope(f"User not found (id:{user_id})")
        return set(user_row.domain.allowed_docker_registries)

    def _get_effective_user_id_in_user_scope(self, ctx: ClientContext, scope: UserScope) -> UUID:
        """
        Determine the effective user ID for permission checks.

        For admin-level users (ADMIN, SUPERADMIN), return the user_id from the UserScope object,
        for allowing them to access customized images of the users.
        For regular users, return the user_id from the client context for permission check.
        """
        if ctx.user_role in [UserRole.ADMIN, UserRole.SUPERADMIN]:
            return scope.user_id
        return ctx.user_id

    async def _in_user_scope(
        self,
        ctx: ClientContext,
        scope: UserScope,
    ) -> ImagePermissionContext:
        allowed_registries = await self._get_allowed_registries_for_user(ctx, scope.user_id)

        permissions = await self.calculate_permission(ctx, scope)
        image_id_permission_map: dict[UUID, frozenset[ImagePermission]] = {}

        img_query_stmt = (
            sa.select(ImageRow)
            .join(ImageRow.registry_row)
            .options(load_only(ImageRow.id, ImageRow.labels, ImageRow.registry))
            .where(ContainerRegistryRow.is_global == true())
        )

        access_criteria = ImageAccessCriteria(
            user_id=self._get_effective_user_id_in_user_scope(ctx, scope),
            allowed_registries=allowed_registries,
        )

        for row in await self.db_session.scalars(img_query_stmt):
            image_row = cast(ImageRow, row)
            if not image_row.customized or not access_criteria.is_accessible_image(image_row):
                continue

            image_id_permission_map[image_row.id] = permissions

        return ImagePermissionContext(
            object_id_to_additional_permission_map=image_id_permission_map
        )

    async def _in_domain_scope(
        self,
        ctx: ClientContext,
        scope: DomainScope,
    ) -> ImagePermissionContext:
        from .container_registry import ContainerRegistryRow
        from .domain import DomainRow

        permissions = await self.calculate_permission(ctx, scope)
        image_id_permission_map: dict[UUID, frozenset[ImagePermission]] = {}

        domain_query_stmt = sa.select(DomainRow).where(DomainRow.name == scope.domain_name)
        domain_row = cast(Optional[DomainRow], await self.db_session.scalar(domain_query_stmt))
        if domain_row is None:
            raise InvalidScope(f"Domain not found (n:{scope.domain_name})")

        allowed_registries: set[str] = set(domain_row.allowed_docker_registries)

        img_query_stmt = (
            sa.select(ImageRow)
            .join(ImageRow.registry_row)
            .options(load_only(ImageRow.id, ImageRow.registry, ImageRow.labels))
            .where(ContainerRegistryRow.is_global == true())
        )

        access_criteria = ImageAccessCriteria(
            user_id=ctx.user_id,
            allowed_registries=allowed_registries,
        )

        for row in await self.db_session.scalars(img_query_stmt):
            image_row = cast(ImageRow, row)
            if not access_criteria.is_accessible_image(image_row):
                continue

            image_id_permission_map[image_row.id] = permissions

        return ImagePermissionContext(
            object_id_to_additional_permission_map=image_id_permission_map
        )

    async def _get_user_accessible_project_scopes(
        self,
        ctx: ClientContext,
        scope: UserScope,
    ) -> list[ProjectScope]:
        from .group import AssocGroupUserRow

        get_assoc_group_ids_stmt = sa.select(AssocGroupUserRow.group_id).where(
            AssocGroupUserRow.user_id == scope.user_id
        )
        group_ids = await self.db_session.scalars(get_assoc_group_ids_stmt)

        return [ProjectScope(project_id=group_id) for group_id in group_ids]

    async def _get_domain_accessible_project_scopes(
        self,
        ctx: ClientContext,
        scope: DomainScope,
    ) -> list[ProjectScope]:
        from .group import GroupRow

        stmt = sa.select(GroupRow.id).where(GroupRow.domain_name == scope.domain_name)
        project_ids = await self.db_session.scalars(stmt)
        return [ProjectScope(project_id=proj_id) for proj_id in project_ids]

    async def _verify_project_scope_and_calculate_permission(
        self, ctx: ClientContext, scope: ProjectScope
    ) -> frozenset[ImagePermission]:
        from .group import GroupRow

        group_query_stmt = sa.select(GroupRow).where(GroupRow.id == scope.project_id)
        group_row = cast(Optional[GroupRow], await self.db_session.scalar(group_query_stmt))
        if group_row is None:
            raise InvalidScope(f"Project not found (project_id: {scope.project_id})")

        return await self.calculate_permission(ctx, scope)

    async def _in_project_scopes_by_registry_condition(
        self,
        ctx: ClientContext,
        scopes: list[ProjectScope],
        registry_condition_factory: Callable[[list[Any]], Any],
        filter_global_registry: bool = False,
    ) -> ImagePermissionContext:
        from .container_registry import ContainerRegistryRow

        project_ids = [scope.project_id for scope in scopes]
        project_id_to_permission_map: dict[str, frozenset[ImagePermission]] = {}

        for scope in scopes:
            permissions = await self._verify_project_scope_and_calculate_permission(ctx, scope)
            project_id_to_permission_map[str(scope.project_id)] = permissions

        image_select_stmt = (
            sa.select(ImageRow)
            .join(ImageRow.registry_row)
            .options(
                joinedload(ImageRow.registry_row).joinedload(
                    ContainerRegistryRow.association_container_registries_groups_rows
                )
            )
            .where(registry_condition_factory(project_ids))
        )

        image_id_to_permission_map: dict[UUID, frozenset[ImagePermission]] = {}

        result = (await self.db_session.scalars(image_select_stmt)).unique()
        for row in result:
            img_row = cast(ImageRow, row)

            allowed_registries = await self._get_allowed_registries_for_user(ctx, ctx.user_id)
            access_criteria = ImageAccessCriteria(
                user_id=ctx.user_id,
                allowed_registries=allowed_registries,
            )

            if not access_criteria.is_accessible_image(img_row):
                continue

            if filter_global_registry:
                # Assumption: permissions for global registry images is same across all projects.
                image_id_to_permission_map[img_row.id] = list(
                    project_id_to_permission_map.values()
                )[0]
            else:
                assoc_project_ids = [
                    assoc.group_id
                    for assoc in img_row.registry_row.association_container_registries_groups_rows
                ]
                for project_id in assoc_project_ids:
                    image_id_to_permission_map[img_row.id] = project_id_to_permission_map[
                        str(project_id)
                    ]

        return ImagePermissionContext(
            object_id_to_additional_permission_map=image_id_to_permission_map
        )

    async def _in_project_scopes_global(
        self,
        ctx: ClientContext,
        scopes: list[ProjectScope],
    ) -> ImagePermissionContext:
        from .container_registry import ContainerRegistryRow

        def global_registry_condition(project_ids: list[Any]):
            return ContainerRegistryRow.is_global == true()

        return await self._in_project_scopes_by_registry_condition(
            ctx, scopes, global_registry_condition, True
        )

    async def _in_project_scopes_non_global(
        self,
        ctx: ClientContext,
        scopes: list[ProjectScope],
    ) -> ImagePermissionContext:
        from .association_container_registries_groups import AssociationContainerRegistriesGroupsRow
        from .container_registry import ContainerRegistryRow

        def non_global_registry_condition(project_ids: list[Any]):
            return ContainerRegistryRow.association_container_registries_groups_rows.any(
                AssociationContainerRegistriesGroupsRow.group_id.in_(project_ids)
            )

        return await self._in_project_scopes_by_registry_condition(
            ctx, scopes, non_global_registry_condition, False
        )

    @override
    @classmethod
    async def _permission_for_owner(
        cls,
    ) -> frozenset[ImagePermission]:
        return OWNER_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_admin(
        cls,
    ) -> frozenset[ImagePermission]:
        return ADMIN_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_monitor(
        cls,
    ) -> frozenset[ImagePermission]:
        return MONITOR_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_privileged_member(
        cls,
    ) -> frozenset[ImagePermission]:
        return PRIVILEGED_MEMBER_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_member(
        cls,
    ) -> frozenset[ImagePermission]:
        return MEMBER_PERMISSIONS


async def get_permission_ctx(
    db_conn: AsyncConnection,
    ctx: ClientContext,
    target_scope: ScopeType,
    requested_permission: ImagePermission,
) -> ImagePermissionContext:
    async with ctx.db.begin_readonly_session(db_conn) as db_session:
        builder = ImagePermissionContextBuilder(db_session)
        permission_ctx = await builder.build(ctx, target_scope, requested_permission)
    return permission_ctx
